Package: hubEvals
Title: Tools for Scoring and Evaluating Hubverse Model Outputs
Version: 0.3.0
Authors@R: c(
    person(given = "Anna",
           family = "Krystalli",
           email = "annakrystalli@googlemail.com",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-2378-4915")),
    person(given = "Nicholas",
           family = "Reich",
           email = "nick@umass.edu",
           role = c("aut"),
           comment = c(ORCID = "0000-0003-3503-9899")),
    person(given = "Evan",
           family = "Ray",
           role = c("aut")),
    person(given = "Nikos",
           family = "Bosse",
           email = "nikosbosse@gmail.com",
           role = c("aut"),
           comment = c(ORCID = "0000-0002-7750-5280")),
    person(given = "Kimberlyn",
           family = "Roosa",
           email = "kimberlyn626@gmail.com",
           role = c("aut")),
    person(given = "Zhian",
           family = "Kamvar",
           role = c("ctb"),
           comment = c(ORCID = "0000-0003-1458-7108")),
    person(given = "Li",
           family = "Shandross",
           email = "lshandross@umass.edu",
           role = c("ctb"),
           comment = c(ORCID = "0009-0008-1348-1954")),
    person(given = "Becky",
           family = "Sweger",
           role = c("ctb")),
    person(given = "Lucie",
           family = "Contamin",
           role = c("ctb")),
    person("Consortium of Infectious Disease Modeling Hubs",
           role = "cph")
    )
Description: 
    Provides tools for scoring and evaluating 'hubverse' model outputs
    against observed data, wrapping scoring workflows from the
    'scoringutils' package and bridging hubverse model output formats
    to 'scoringutils' forecast classes.
License: MIT + file LICENSE
Encoding: UTF-8
Imports: cli, dplyr, hubUtils (>= 1.2.0), purrr, rlang, scoringutils
        (>= 2.2.0), tibble
RoxygenNote: 7.3.3
URL: https://hubverse-org.github.io/hubEvals/,
        https://github.com/hubverse-org/hubEvals
BugReports: https://github.com/hubverse-org/hubEvals/issues
Additional_repositories: https://hubverse-org.r-universe.dev
Depends: R (>= 4.1.0)
Config/Needs/website: hubverse-org/hubStyle
Suggests: hubExamples, knitr, rmarkdown, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-23 08:28:43 UTC; Anna
Author: Anna Krystalli [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2378-4915>),
  Nicholas Reich [aut] (ORCID: <https://orcid.org/0000-0003-3503-9899>),
  Evan Ray [aut],
  Nikos Bosse [aut] (ORCID: <https://orcid.org/0000-0002-7750-5280>),
  Kimberlyn Roosa [aut],
  Zhian Kamvar [ctb] (ORCID: <https://orcid.org/0000-0003-1458-7108>),
  Li Shandross [ctb] (ORCID: <https://orcid.org/0009-0008-1348-1954>),
  Becky Sweger [ctb],
  Lucie Contamin [ctb],
  Consortium of Infectious Disease Modeling Hubs [cph]
Maintainer: Anna Krystalli <annakrystalli@googlemail.com>
Repository: CRAN
Date/Publication: 2026-06-29 13:40:02 UTC
Built: R 4.5.2; ; 2026-06-29 16:54:35 UTC; unix

# Generated by roxygen2: do not edit by hand

export(score_model_out)
export(transform_pmf_model_out)
export(transform_point_model_out)
export(transform_quantile_model_out)
export(transform_sample_model_out)
importFrom(rlang,.data)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# hubEvals <a href="https://hubverse-org.github.io/hubEvals/"><img src="man/figures/logo.png" align="right" height="131" alt="hubEvals website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/hubEvals)](https://CRAN.R-project.org/package=hubEvals)
[![R-CMD-check](https://github.com/hubverse-org/hubEvals/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/hubverse-org/hubEvals/actions/workflows/R-CMD-check.yaml)

<!-- badges: end -->

The goal of hubEvals is to provide tools for evaluating infectious
disease model outputs. This package is part of the
[Hubverse](https://hubverse.io) project, which aims to provide a suite
of tools for infectious disease modeling hubs.

## Installation

### Latest

You can install the [latest version of hubEvals from the
R-universe](https://hubverse-org.r-universe.dev/hubEvals):

``` r
install.packages("hubEvals", repos = c("https://hubverse-org.r-universe.dev", "https://cloud.r-project.org"))
```

### Development

If you want to test out new features that have not yet been released,
you can install the development version of hubEvals from
[GitHub](https://github.com/) with:

``` r
# install.packages("remotes")
remotes::install_github("hubverse-org/hubEvals")
```

## Example

Predictions can be evaluated directly using the scoring function in
`hubEvals`, which assumes a hubverse format for the model outputs and
target data:

``` r
library(hubEvals)

# compute default metrics (in this case, absolute error) for
# median forecasts, summarized by the mean score for each model
median_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(output_type == "median"), # only one output type allowed
  oracle_output = hubExamples::forecast_oracle_output,
  by = "model_id"
)
median_scores
#>             model_id ae_point
#>               <char>    <num>
#> 1: Flusight-baseline  401.875
#> 2:   MOBS-GLEAM_FLUH  416.375
#> 3:          PSI-DICE  277.000

# compute WIS and interval coverage rates at 80% and 90% levels based on
# quantile forecasts, summarized by the mean score for each model
quantile_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(output_type == "quantile"), # only one output type allowed
  oracle_output = hubExamples::forecast_oracle_output,
  metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
  relative_metrics = "wis",
  by = "model_id"
)
quantile_scores
#> Key: <model_id>
#>             model_id      wis interval_coverage_80 interval_coverage_90 wis_relative_skill
#>               <char>    <num>                <num>                <num>              <num>
#> 1: Flusight-baseline 329.4545                  0.0               0.1250          1.1473659
#> 2:   MOBS-GLEAM_FLUH 315.2393                  0.5               0.5625          1.0978597
#> 3:          PSI-DICE 227.9527                  0.5               0.5000          0.7938733

# compute log scores based on pmf predictions for categorical targets,
# summarized by the mean score for each combination of model and location.
# Note: if the model_out_tbl had forecasts for multiple targets using a
# pmf output_type with different bins, it would be necessary to score the
# predictions for those targets separately.
pmf_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(output_type == "pmf"), # only one output type allowed
  oracle_output = hubExamples::forecast_oracle_output,
  metrics = c("log_score", "rps"),
  by = c("model_id", "location", "horizon"),
  output_type_id_order = c("low", "moderate", "high", "very high")
)
head(pmf_scores)
#>             model_id location horizon   log_score          rps
#>               <char>   <char>   <int>       <num>        <num>
#> 1: Flusight-baseline       25       0  0.02107606 0.0008531043
#> 2: Flusight-baseline       25       1  6.69652380 0.5029240066
#> 3: Flusight-baseline       25       2 17.73313203 1.0057355863
#> 4: Flusight-baseline       25       3         Inf 1.8665126816
#> 5: Flusight-baseline       48       0  2.18418007 0.4873966597
#> 6: Flusight-baseline       48       1  7.49960792 0.9659026096
```

[Sample
forecasts](https://docs.hubverse.io/en/latest/user-guide/sample-output-type.html)
can be scored marginally (each modeling task scored independently) or
jointly using compound scoring:

``` r
# marginal sample scoring with CRPS
sample_scores <- hubExamples::forecast_outputs |>
  dplyr::filter(output_type == "sample") |>
  score_model_out(
    oracle_output = hubExamples::forecast_oracle_output,
    metrics = "crps",
    by = "model_id"
  )
sample_scores
#>             model_id     crps
#>               <char>    <num>
#> 1: Flusight-baseline 351.5887
#> 2:   MOBS-GLEAM_FLUH 347.1502
#> 3:          PSI-DICE 247.3640
```

[Compound
scoring](https://docs.hubverse.io/en/latest/user-guide/sample-output-type.html#compound-modeling-tasks)
uses the energy score to evaluate the joint distribution across task
dimensions that vary within a sample draw. The `compound_taskid_set`
specifies which task IDs stay constant within a sample group and can be
found by referencing the hub’s `tasks.json` configuration file. Here,
each draw spans all horizons for a given reference date and location
(i.e., a trajectory over time).

``` r
compound_scores <- hubExamples::forecast_outputs |>
  dplyr::filter(output_type == "sample") |>
  score_model_out(
    oracle_output = hubExamples::forecast_oracle_output,
    compound_taskid_set = c("reference_date", "location"),
    by = "model_id"
  )
compound_scores
#>             model_id energy_score variogram_score
#>               <char>        <num>           <num>
#> 1: Flusight-baseline     772.7608        1524.474
#> 2:   MOBS-GLEAM_FLUH     811.4625        1695.037
#> 3:          PSI-DICE     571.0879        1264.238
```

Or, users may transform predictions into a `forecast` object that can be
used as an input to `scoringutils` functions and use their tooling
directly.

``` r
median_forecast <- transform_point_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(output_type == "median"),
  oracle_output = hubExamples::forecast_oracle_output,
  output_type = "median"
)
median_forecast

quantile_forecast <- transform_quantile_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(output_type == "quantile"),
  oracle_output = hubExamples::forecast_oracle_output
)
quantile_forecast

pmf_forecasts <- transform_pmf_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(output_type == "pmf"),
  oracle_output = hubExamples::forecast_oracle_output,
  output_type_id_order = c("low", "moderate", "high", "very high")
)
pmf_forecasts

sample_forecast <- hubExamples::forecast_outputs |>
  dplyr::filter(output_type == "sample") |>
  transform_sample_model_out(
    oracle_output = hubExamples::forecast_oracle_output
  )
sample_forecast
```

## Code of Conduct

Please note that the hubEvals package is released with a [Contributor
Code of
Conduct](https://github.com/hubverse-org/hubEvals/blob/main/.github/CODE_OF_CONDUCT.md).
By contributing to this project, you agree to abide by its terms.

## Contributing

Interested in contributing back to the open-source Hubverse project?
Learn more about how to [get involved in the Hubverse
Community](https://hubverse.io/community/) or [how to contribute to the
hubEvals
package](https://github.com/hubverse-org/hubEvals/blob/main/.github/CONTRIBUTING.md).

#' Transform pmf model output into a forecast object
#'
#' @param model_out_tbl Model output tibble with predictions
#' @param oracle_output Predictions that would have been generated by an oracle
#' model that knew the observed target data values in advance
#' @param output_type_id_order For ordinal variables in pmf format, this is a
#' vector of levels for pmf forecasts, in increasing order of the levels. The
#' order of the values for the output_type_id can be found by referencing the
#' hub's tasks.json configuration file. For all output types other than pmf,
#' this is ignored.
#'
#' @return A `forecast_ordinal` object (when `output_type_id_order` is provided)
#'   or a `forecast_nominal` object (when `output_type_id_order` is `NULL`).
#'
#' @examplesIf requireNamespace("hubExamples", quietly = TRUE)
#' # Nominal pmf forecast (no output_type_id_order provided)
#' pmf_outputs <- hubExamples::forecast_outputs |>
#'   dplyr::filter(.data[["output_type"]] == "pmf")
#'
#' nominal_forecast <- transform_pmf_model_out(
#'   model_out_tbl = pmf_outputs,
#'   oracle_output = hubExamples::forecast_oracle_output
#' )
#' nominal_forecast
#'
#' # Ordinal pmf forecast (output_type_id_order provided)
#' ordinal_forecast <- transform_pmf_model_out(
#'   model_out_tbl = pmf_outputs,
#'   oracle_output = hubExamples::forecast_oracle_output,
#'   output_type_id_order = c("low", "moderate", "high", "very high")
#' )
#' ordinal_forecast
#'
#' @importFrom rlang .data
#' @export
transform_pmf_model_out <- function(
  model_out_tbl,
  oracle_output,
  output_type_id_order = NULL
) {
  model_out_tbl <- validate_model_oracle_out(model_out_tbl, oracle_output)

  # subset both model_out_tbl and oracle_output to output_type == "pmf"
  model_out_tbl <- model_out_tbl |>
    dplyr::filter(.data[["output_type"]] == "pmf")

  if (c("output_type") %in% colnames(oracle_output)) {
    oracle_output <- oracle_output |>
      dplyr::filter(.data[["output_type"]] == "pmf") |>
      dplyr::select(-c("output_type"))
  }

  # validate or set output_type_id_order
  if (!is.null(output_type_id_order)) {
    output_type_id_order <- validate_output_type_id_order(
      output_type_id_order,
      model_out_tbl
    )
    is_ordinal <- TRUE
  } else {
    is_ordinal <- FALSE
    output_type_id_order <- unique(model_out_tbl$output_type_id)
  }

  task_id_cols <- get_task_id_cols(model_out_tbl)

  # assemble data in scoringutils format
  model_out_tbl <- model_out_tbl |>
    dplyr::rename(model = "model_id")

  data <- dplyr::left_join(
    model_out_tbl,
    oracle_output,
    by = c(
      task_id_cols[task_id_cols %in% colnames(oracle_output)],
      "output_type_id"
    ),
    relationship = "many-to-one"
  ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(c(task_id_cols, "model")))) |>
    dplyr::mutate(
      observation = .data[["output_type_id"]][.data[["oracle_value"]] == 1],
      observation = factor(
        .data[["observation"]],
        levels = output_type_id_order,
        ordered = is_ordinal
      ),
      output_type_id = factor(
        .data[["output_type_id"]],
        levels = output_type_id_order,
        ordered = is_ordinal
      )
    ) |>
    dplyr::ungroup() |>
    dplyr::select(-dplyr::all_of("oracle_value"))

  if (is_ordinal) {
    forecast_pmf <- scoringutils::as_forecast_ordinal(
      data,
      forecast_unit = c("model", task_id_cols),
      observed = "observation",
      predicted = "value",
      predicted_label = "output_type_id"
    )
  } else {
    forecast_pmf <- scoringutils::as_forecast_nominal(
      data,
      forecast_unit = c("model", task_id_cols),
      observed = "observation",
      predicted = "value",
      predicted_label = "output_type_id"
    )
  }

  forecast_pmf
}


#' Validate `output_type_id_order` for ordinal variables:
#' - Must be a vector
#' - Must be (set-)equal to the set of all unique `output_type_id` values in `model_out_tbl`
#'
#' This function expects that the model_out_tbl has already been filtered to
#' output_type == "pmf". This is done in `transform_pmf_model_out`.
#'
#' @noRd
validate_output_type_id_order <- function(output_type_id_order, model_out_tbl) {
  if (!is.vector(output_type_id_order)) {
    cli::cli_abort("`output_type_id_order` must be a vector.")
  }

  present_levels <- unique(model_out_tbl$output_type_id)
  extra_order_levels <- setdiff(output_type_id_order, present_levels)
  missing_order_levels <- setdiff(present_levels, output_type_id_order)
  if (length(extra_order_levels) > 0 || length(missing_order_levels) > 0) {
    cli::cli_abort(
      c(
        "x" = "`output_type_id_order` must align with the set of all unique
               `output_type_id` values in `model_out_tbl`.",
        "!" = ifelse(
          length(extra_order_levels) == 0,
          NULL,
          "The following levels were present in `output_type_id_order` but not in `model_out_tbl`:
          {.val {extra_order_levels}}."
        ),
        "!" = ifelse(
          length(missing_order_levels) == 0,
          NULL,
          "The following levels were present in `model_out_tbl` but not in `output_type_id_order`:
          {.val {missing_order_levels}};"
        )
      )
    )
  }

  output_type_id_order
}

# This file uses test fixture data that is created in
# tests/testthat/testdata/make_pmf_test_data.R, and can be loaded by the calls
#
# model_out_tbl <- pmf_test_model_out_tbl()
# oracle_output <- pmf_test_oracle_output()
# exp_forecast <- pmf_test_exp_forecast()
#
# By construction, model_out_tbl and oracle_output are valid inputs to transform_pmf_model_out,
# and exp_forecast is the expected return value from transform_pmf_model_out

test_that("transform_pmf_model_out succeeds with valid inputs -- nominal", {
  model_out_tbl <- pmf_test_model_out_tbl()
  oracle_output <- pmf_test_oracle_output()
  exp_forecast <- pmf_test_exp_forecast()

  act_forecast <- transform_pmf_model_out(
    model_out_tbl = model_out_tbl,
    oracle_output = oracle_output
  )

  expect_equal(act_forecast, exp_forecast)
})


test_that("transform_pmf_model_out succeeds with valid inputs -- ordinal", {
  model_out_tbl <- pmf_test_model_out_tbl()
  oracle_output <- pmf_test_oracle_output()
  exp_forecast <- pmf_test_exp_forecast() |>
    dplyr::mutate(
      predicted_label = factor(
        predicted_label,
        levels = c("cat", "dog", "bird"),
        ordered = TRUE
      ),
      observed = factor(
        observed,
        levels = c("cat", "dog", "bird"),
        ordered = TRUE
      )
    )
  class(exp_forecast) <- c(
    "forecast_ordinal",
    "forecast",
    "data.table",
    "data.frame"
  )

  act_forecast <- transform_pmf_model_out(
    model_out_tbl = model_out_tbl,
    oracle_output = oracle_output,
    output_type_id_order = c("cat", "dog", "bird")
  )

  expect_equal(act_forecast, exp_forecast)
})


test_that("transform_pmf_model_out throws an error with invalid output_type_id_order", {
  model_out_tbl <- pmf_test_model_out_tbl()
  oracle_output <- pmf_test_oracle_output()

  expect_error(
    act_forecast <- transform_pmf_model_out(
      model_out_tbl = model_out_tbl,
      oracle_output = oracle_output,
      output_type_id_order = c("cat", "bird", "platypus"),
      "`output_type_id_order` must align with the set of all unique `output_type_id` values in `model_out_tbl`."
    )
  )
})


test_that("transform_pmf_model_out doesn't depend on specific column names for task id variables", {
  model_out_tbl <- pmf_test_model_out_tbl() |>
    dplyr::rename(loc = location, date = target_date)
  oracle_output <- pmf_test_oracle_output() |>
    dplyr::rename(loc = location, date = target_date)
  exp_forecast <- pmf_test_exp_forecast() |>
    dplyr::rename(loc = location, date = target_date)

  act_forecast <- transform_pmf_model_out(
    model_out_tbl = model_out_tbl,
    oracle_output = oracle_output
  )

  expect_equal(act_forecast, exp_forecast)
})


test_that("transform_pmf_model_out throws an error if model_out_tbl has no rows", {
  # Error is thrown by checkmate::assert_data_frame() via scoringutils::assert_forecast_generic()
  model_out_tbl <- pmf_test_model_out_tbl()
  oracle_output <- pmf_test_oracle_output()
  expect_error(
    suppressWarnings(transform_pmf_model_out(
      model_out_tbl = model_out_tbl[0, ],
      oracle_output = oracle_output
    )),
    regexp = "Must have at least 1 rows, but has 0 rows."
  )
})


test_that("many-to-one relationship exists between model_out_tbl and oracle_output", {
  model_out_tbl <- pmf_test_model_out_tbl()
  oracle_output <- pmf_test_oracle_output() |>
    dplyr::select(-location)

  expect_error(
    suppressMessages(transform_pmf_model_out(
      model_out_tbl = model_out_tbl,
      oracle_output = oracle_output
    )),
    regexp = "Each row in `x` must match at most 1 row in `y`."
  )
})

# Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
forecast_outputs <- hubExamples::forecast_outputs
forecast_oracle_output <- hubExamples::forecast_oracle_output

# expected scores
## log scores
exp_log_scores <- forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "pmf") |>
  dplyr::left_join(
    forecast_oracle_output |>
      dplyr::filter(.data[["output_type"]] == "pmf") |>
      dplyr::select(-dplyr::all_of(c("output_type"))),
    by = c("location", "target_end_date", "target", "output_type_id")
  ) |>
  dplyr::filter(.data[["oracle_value"]] == 1) |>
  dplyr::mutate(
    log_score = -1 * log(.data[["value"]])
  ) |>
  dplyr::group_by(dplyr::across(dplyr::all_of(
    c("model_id", "location")
  ))) |>
  dplyr::summarize(
    log_score = mean(.data[["log_score"]]),
    .groups = "drop"
  )

## rps scores
## see Eq (1) of Weigen et al. (2006) The Discrete Brier and Ranked Probability Skill Scores.
## Monthly Weather Review, 135, 118-124.
exp_rps_scores <- forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "pmf") |>
  dplyr::left_join(
    forecast_oracle_output |>
      dplyr::filter(.data[["output_type"]] == "pmf") |>
      dplyr::select(-dplyr::all_of(c("output_type"))),
    by = c("location", "target_end_date", "target", "output_type_id")
  ) |>
  dplyr::group_by(dplyr::across(dplyr::all_of(
    c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  ))) |>
  dplyr::summarize(
    rps = sum((cumsum(.data[["value"]]) - cumsum(.data[["oracle_value"]]))^2)
  ) |>
  dplyr::group_by(dplyr::across(dplyr::all_of(
    c("model_id", "location")
  ))) |>
  dplyr::summarize(
    rps = mean(.data[["rps"]]),
    .groups = "drop"
  )

exp_scores <- dplyr::full_join(
  exp_log_scores,
  exp_rps_scores,
  by = c("model_id", "location")
)

write.csv(
  exp_scores,
  testthat::test_path("testdata", "exp_pmf_scores.csv"),
  row.names = FALSE
)

#' Transform sample model output into a forecast object
#'
#' @param model_out_tbl Model output tibble with predictions
#' @param oracle_output Predictions that would have been generated by an oracle
#' model that knew the observed target data values in advance
#' @param compound_taskid_set Character vector of task ID column names that
#'   stay constant within each sample draw (i.e., define the compound modeling
#'   task grouping). When `NULL` (the default), each modeling task is scored
#'   independently (marginal scoring). When provided, sample draws are treated
#'   as joint predictions across the task ID dimensions **not** in
#'   `compound_taskid_set`, and multivariate scoring metrics are used.
#'
#' @return A `forecast_sample` object (when `compound_taskid_set` is `NULL`)
#'   or a `forecast_multivariate_sample` object (when `compound_taskid_set` is
#'   provided).
#'
#' @examplesIf requireNamespace("hubExamples", quietly = TRUE)
#' # Marginal sample forecast: each modeling task scored independently
#' sample_forecast <- hubExamples::forecast_outputs |>
#'   dplyr::filter(.data[["output_type"]] == "sample") |>
#'   transform_sample_model_out(
#'     oracle_output = hubExamples::forecast_oracle_output
#'   )
#' sample_forecast
#'
#' # Compound sample forecast: jointly score across non-compound task IDs
#' compound_forecast <- hubExamples::forecast_outputs |>
#'   dplyr::filter(.data[["output_type"]] == "sample") |>
#'   transform_sample_model_out(
#'     oracle_output = hubExamples::forecast_oracle_output,
#'     compound_taskid_set = c("reference_date", "location")
#'   )
#' compound_forecast
#'
#' @importFrom rlang .data
#' @export
transform_sample_model_out <- function(
  model_out_tbl,
  oracle_output,
  compound_taskid_set = NULL
) {
  model_out_tbl <- validate_model_oracle_out(model_out_tbl, oracle_output)

  task_id_cols <- get_task_id_cols(model_out_tbl)

  model_out_tbl <- model_out_tbl |>
    dplyr::filter(.data[["output_type"]] == "sample") |>
    dplyr::rename(model = "model_id")

  # Filter oracle_output to sample rows only when it still carries the
  # `output_type` column (i.e. mixed output types).
  if ("output_type" %in% colnames(oracle_output)) {
    oracle_output <- dplyr::filter(
      oracle_output,
      .data[["output_type"]] == "sample"
    )
  }
  # Always drop output_type/output_type_id if present: output_type_id can
  # remain after a user-side filter that removed output_type, and oracle
  # output has no sample id to join on.
  oracle_output <- dplyr::select(
    oracle_output,
    -dplyr::any_of(c("output_type", "output_type_id"))
  )

  data <- dplyr::left_join(
    model_out_tbl,
    oracle_output,
    by = task_id_cols[task_id_cols %in% colnames(oracle_output)],
    relationship = "many-to-one"
  )

  if (is.null(compound_taskid_set)) {
    # Marginal: score each modeling task independently
    forecast <- scoringutils::as_forecast_sample(
      data,
      forecast_unit = c("model", task_id_cols),
      observed = "oracle_value",
      predicted = "value",
      sample_id = "output_type_id"
    )
  } else {
    validate_compound_taskid_set(data, compound_taskid_set, task_id_cols)

    # Compound: score joint distribution across non-compound task IDs.
    # The non-compound task IDs are the dimensions that vary within a single
    # sample draw -- these map to scoringutils' `joint_across` parameter.
    joint_across <- setdiff(task_id_cols, compound_taskid_set)
    forecast <- scoringutils::as_forecast_multivariate_sample(
      data,
      joint_across = joint_across,
      forecast_unit = c("model", task_id_cols),
      observed = "oracle_value",
      predicted = "value",
      sample_id = "output_type_id"
    )
  }

  forecast
}

# --- Marginal sample scoring ---

test_that("hubExamples sample data transforms to forecast_sample for marginal scoring", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  act_forecast <- transform_sample_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output
  )

  expect_s3_class(
    act_forecast,
    c("forecast_sample", "forecast", "data.table", "data.frame")
  )

  # same number of rows as sample predictions
  expect_equal(nrow(act_forecast), nrow(sample_tbl))

  # correct column mapping
  expect_named(
    act_forecast,
    c(
      "sample_id",
      "predicted",
      "observed",
      "model",
      "reference_date",
      "target",
      "horizon",
      "location",
      "target_end_date"
    )
  )

  # predicted values match original values. We join rather than compare
  # positionally because as_forecast_sample() converts to data.table which
  # may reorder rows.
  merged <- dplyr::inner_join(
    act_forecast,
    sample_tbl,
    by = dplyr::join_by(
      model == model_id,
      sample_id == output_type_id,
      reference_date,
      target,
      horizon,
      location,
      target_end_date
    ),
    relationship = "one-to-one"
  )
  expect_equal(merged$predicted, merged$value)

  # observed values correctly merged from oracle output
  oracle_sample <- forecast_oracle_output |>
    dplyr::filter(.data[["output_type"]] == "sample") |>
    dplyr::select(-c("output_type", "output_type_id"))
  obs_merged <- dplyr::inner_join(
    act_forecast,
    oracle_sample,
    by = dplyr::join_by(location, target_end_date, target),
    relationship = "many-to-one"
  )
  expect_equal(obs_merged$observed, obs_merged$oracle_value)
})


test_that("oracle output without output_type/output_type_id columns works", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # Strip output_type and output_type_id from oracle
  oracle_simple <- forecast_oracle_output |>
    dplyr::filter(.data[["output_type"]] == "sample") |>
    dplyr::select(-c("output_type", "output_type_id"))

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  act_forecast <- transform_sample_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = oracle_simple
  )

  expect_s3_class(act_forecast, "forecast_sample")
  expect_equal(nrow(act_forecast), nrow(sample_tbl))

  # observed values correctly merged from oracle output
  merged <- dplyr::inner_join(
    act_forecast,
    oracle_simple,
    by = dplyr::join_by(location, target_end_date, target),
    relationship = "many-to-one"
  )
  expect_equal(merged$observed, merged$oracle_value)
})


test_that("transform_sample_model_out errors on empty data", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  expect_error(
    suppressWarnings(transform_sample_model_out(
      model_out_tbl = sample_tbl[0, ],
      oracle_output = forecast_oracle_output
    )),
    regexp = "Must have at least 1 rows, but has 0 rows."
  )
})


test_that("transform_sample_model_out errors on missing required columns", {
  skip_if_not_installed("hubExamples")
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  bad_tbl <- data.frame(
    model_id = "A",
    output_type = "sample",
    value = 1
  )

  expect_error(
    transform_sample_model_out(
      model_out_tbl = bad_tbl,
      oracle_output = forecast_oracle_output
    ),
    regexp = "does not contain required columns"
  )
})


# --- Compound sample scoring ---

test_that("compound_taskid_set produces forecast_sample_multivariate", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  act_forecast <- transform_sample_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output,
    compound_taskid_set = c("reference_date", "location")
  )

  expect_s3_class(
    act_forecast,
    c("forecast_sample_multivariate", "forecast", "data.table", "data.frame")
  )
  expect_equal(nrow(act_forecast), nrow(sample_tbl))
  expect_named(
    act_forecast,
    c(
      "sample_id",
      "predicted",
      "observed",
      "model",
      "reference_date",
      "target",
      "horizon",
      "location",
      "target_end_date",
      ".mv_group_id"
    )
  )
})


test_that("validate_compound_taskid_set errors on non-character input", {
  expect_error(
    validate_compound_taskid_set(
      data = data.frame(),
      compound_taskid_set = 123,
      task_id_cols = c("location", "horizon")
    ),
    regexp = "must be a non-empty character vector"
  )
})


test_that("validate_compound_taskid_set errors on empty vector", {
  expect_error(
    validate_compound_taskid_set(
      data = data.frame(),
      compound_taskid_set = character(0),
      task_id_cols = c("location", "horizon")
    ),
    regexp = "must be a non-empty character vector"
  )
})


test_that("validate_compound_taskid_set errors on invalid column names", {
  expect_error(
    validate_compound_taskid_set(
      data = data.frame(),
      compound_taskid_set = c("location", "nonexistent"),
      task_id_cols = c("location", "horizon")
    ),
    regexp = "contains invalid column"
  )
})


test_that("validate_compound_taskid_set errors when all task IDs included", {
  expect_error(
    validate_compound_taskid_set(
      data = data.frame(),
      compound_taskid_set = c("location", "horizon"),
      task_id_cols = c("location", "horizon")
    ),
    regexp = "includes all task ID columns"
  )
})


test_that("validate_compound_taskid_set errors when joint_across dims don't vary", {
  # Create data where non-compound task IDs (horizon) don't vary within draws.
  # A draw is identified by compound_taskid_set + output_type_id.
  data <- data.frame(
    output_type_id = c("s1", "s2"),
    location = c("25", "25"),
    horizon = c(0, 0)
  )

  expect_error(
    validate_compound_taskid_set(
      data = data,
      compound_taskid_set = "location",
      task_id_cols = c("location", "horizon")
    ),
    regexp = "must vary within sample draws"
  )
})


test_that("finer compound_taskid_set with unique IDs errors", {
  # Marginal data with globally unique sample IDs: each sample covers only
  # one (location, horizon) combination. When IDs are unique, joint_across
  # dimensions don't vary within draws, so validation correctly errors.
  data <- data.frame(
    output_type_id = as.character(1:12),
    location = rep(c("US", "UK"), each = 6),
    horizon = rep(rep(1:3, each = 2), 2)
  )

  expect_error(
    validate_compound_taskid_set(
      data = data,
      compound_taskid_set = "location",
      task_id_cols = c("location", "horizon")
    ),
    regexp = "must vary within sample draws"
  )
})


test_that("coarser compound_taskid_set than actual structure passes", {
  skip_if_not_installed("hubExamples")
  # The hubExamples data has compound_taskid_set = {reference_date, location}
  # per tasks.json: each draw has fixed ref_date + location, varying over
  # horizon. Scoring with just {reference_date} is coarser -- it treats
  # location as a joint_across dimension. This passes validation because
  # location does vary within draws (due to sample ID reuse across locations).
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  forecast <- transform_sample_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output,
    compound_taskid_set = "reference_date"
  )

  expect_s3_class(forecast, "forecast_multivariate_sample")
  expect_equal(nrow(forecast), nrow(sample_tbl))

  # The same sample_id appears in multiple .mv_group_ids because the actual
  # draw structure is finer than the specified compound_taskid_set -- each
  # draw is being split across locations.
  sample_groups <- unique(as.data.frame(forecast)[, c(
    "sample_id",
    ".mv_group_id"
  )])
  ids_in_multiple <- sample_groups$sample_id[duplicated(
    sample_groups$sample_id
  )]
  expect_true(length(ids_in_multiple) > 0)
})

#' Transform either mean or median model output into a point forecast object:
#'
#' @param model_out_tbl Model output tibble with predictions
#' @param oracle_output Predictions that would have been generated by an oracle
#' model that knew the observed target data values in advance
#' @param output_type Forecast output type: "mean" or "median"
#'
#' @return forecast_point
#'
#' @details This function transforms a model output tibble in the Hubverse
#' format (with either "mean" or "median" output type) to a scoringutils "point"
#' forecast object
#'
#' @examplesIf requireNamespace("hubExamples", quietly = TRUE)
#' # Median point forecast
#' median_forecast <- hubExamples::forecast_outputs |>
#'   dplyr::filter(.data[["output_type"]] == "median") |>
#'   transform_point_model_out(
#'     oracle_output = hubExamples::forecast_oracle_output,
#'     output_type = "median"
#'   )
#' median_forecast
#'
#' # Mean point forecast
#' mean_forecast <- hubExamples::forecast_outputs |>
#'   dplyr::filter(.data[["output_type"]] == "mean") |>
#'   transform_point_model_out(
#'     oracle_output = hubExamples::forecast_oracle_output,
#'     output_type = "mean"
#'   )
#' mean_forecast
#'
#' @importFrom rlang .data
#' @export
transform_point_model_out <- function(
  model_out_tbl,
  oracle_output,
  output_type
) {
  if (
    (!inherits(output_type, "character")) ||
      (!output_type %in% c("mean", "median"))
  ) {
    cli::cli_abort(
      "invalid 'output_type': {.val {output_type}} Must be 'mean' or 'median'"
    )
  }

  model_out_tbl <- validate_model_oracle_out(model_out_tbl, oracle_output)

  task_id_cols <- get_task_id_cols(model_out_tbl)
  type <- output_type

  model_out_tbl <- model_out_tbl |>
    dplyr::filter(output_type == type) |>
    dplyr::rename(model = "model_id")

  # Filter oracle_output to `type` rows only when it still carries the
  # `output_type` column (i.e. mixed output types).
  if ("output_type" %in% colnames(oracle_output)) {
    oracle_output <- dplyr::filter(oracle_output, output_type == type)
  }
  # Always drop output_type/output_type_id if present: output_type_id can
  # remain after a user-side filter that removed output_type, and neither is
  # used for the point join.
  oracle_output <- dplyr::select(
    oracle_output,
    -dplyr::any_of(c("output_type", "output_type_id"))
  )

  data <- dplyr::left_join(
    model_out_tbl,
    oracle_output,
    by = task_id_cols[task_id_cols %in% colnames(oracle_output)],
    relationship = "many-to-one"
  )

  forecast_point <- scoringutils::as_forecast_point(
    data,
    forecast_unit = c("model", task_id_cols),
    observed = "oracle_value",
    predicted = "value"
  )

  forecast_point
}

test_that("inputs are valid", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  expect_error(
    suppressMessages(transform_point_model_out(
      model_out_tbl = model_out_tbl_1,
      oracle_output = data.frame(),
      output_type = NULL
    )),
    regexp = "invalid 'output_type': Must be 'mean' or 'median'"
  )
  expect_error(
    suppressMessages(transform_point_model_out(
      model_out_tbl = model_out_tbl_1,
      oracle_output = data.frame(),
      output_type = 1
    )),
    regexp = "invalid 'output_type': 1 Must be 'mean' or 'median'"
  )
})


test_that("model_out_tbl_1 output is valid", {
  # Task IDs: location, reference_date, target_end_date, target

  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(oracle_value = observation)
  act_forecast <- transform_point_model_out(
    model_out_tbl = model_out_tbl_1,
    oracle_output = oracle_output_1,
    output_type = "mean"
  )
  exp_forecast <- utils::read.csv(
    test_path("testdata/exp_forecast_1.csv")
  )
  class(exp_forecast) <- c(
    "forecast_point",
    "forecast",
    "data.table",
    "data.frame"
  )
  expect_equal(act_forecast, exp_forecast)
})


test_that("transform_point_model_out() throws error for unexpected columns", {
  # Task IDs: location, reference_date, target_end_date, target

  model_out_tbl_1 <- utils::read.csv(test_path(
    "testdata/model_out_tbl_point_1.csv"
  )) |>
    dplyr::rename(loc = location, trgt = target, date = target_end_date)

  oracle_output_1 <- utils::read.csv(test_path("testdata/target_data_1.csv")) |>
    dplyr::rename(date = target_end_date) # leave location column unchanged

  expect_error(
    suppressMessages(transform_point_model_out(
      model_out_tbl = model_out_tbl_1,
      oracle_output = oracle_output_1,
      output_type = "mean"
    )),
    regexp = "unexpected column"
  )
})

test_that("transform_point_model_out() works with modified column names", {
  # Task IDs: location, reference_date, target_end_date, target
  model_out_tbl_1 <- utils::read.csv(test_path(
    "testdata/model_out_tbl_point_1.csv"
  )) |>
    dplyr::rename(loc = location, trgt = target, date = target_end_date)

  oracle_output_1 <- utils::read.csv(test_path("testdata/target_data_1.csv")) |>
    dplyr::rename(
      loc = location,
      date = target_end_date,
      oracle_value = observation
    )

  act_forecast <- transform_point_model_out(
    model_out_tbl = model_out_tbl_1,
    oracle_output = oracle_output_1,
    output_type = "mean"
  )
  exp_forecast <- utils::read.csv(test_path("testdata/exp_forecast_1.csv")) |>
    dplyr::rename(loc = location, trgt = target, date = target_end_date)
  class(exp_forecast) <- c(
    "forecast_point",
    "forecast",
    "data.table",
    "data.frame"
  )
  expect_equal(act_forecast, exp_forecast, ignore_attr = "class")
})

test_that("Errors is thrown when model_out_tbl_1 has no rows", {
  # Error is thrown by checkmate::assert_data_frame() via scoringutils::assert_forecast_generic()
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(oracle_value = observation)
  expect_error(
    suppressWarnings(transform_point_model_out(
      model_out_tbl = model_out_tbl_1[0, ],
      oracle_output = oracle_output_1,
      output_type = "mean"
    )),
    regexp = "Must have at least 1 rows, but has 0 rows."
  )
})


test_that("many-to-one relationship exists between model_out_tbl and oracle_output", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_2 <- utils::read.csv(
    test_path("testdata/target_data_2.csv")
  ) |>
    dplyr::rename(oracle_value = observation)

  expect_error(
    suppressMessages(transform_point_model_out(
      model_out_tbl = model_out_tbl_1,
      oracle_output = oracle_output_2,
      output_type = "mean"
    )),
    regexp = "Each row in `x` must match at most 1 row in `y`."
  )
})

test_that("hubExamples data set is transformed correctly", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output
  act_forecast <- transform_point_model_out(
    model_out_tbl = forecast_outputs,
    oracle_output = forecast_oracle_output,
    output_type = "mean"
  )

  exp_forecast <- forecast_outputs |>
    dplyr::filter(output_type == "mean") |>
    dplyr::rename(model = model_id)

  expect_s3_class(
    act_forecast,
    c("forecast_point", "forecast", "data.table", "data.frame")
  )
  # same number of row as mean predictions from model_out_tbl,
  # same predicted values
  expect_equal(nrow(exp_forecast), nrow(act_forecast))
  exp_act_forecast <- dplyr::full_join(
    exp_forecast,
    act_forecast,
    by = c(
      "model",
      "reference_date",
      "target",
      "horizon",
      "location",
      "target_end_date"
    )
  )
  expect_equal(exp_act_forecast$predicted, exp_act_forecast$value)

  # correct observed values, in alignment with oracle_output
  exp_act_forecast <- dplyr::left_join(
    act_forecast,
    forecast_oracle_output |> dplyr::filter(output_type == "mean"),
    by = c("target", "location", "target_end_date")
  )
  expect_equal(exp_act_forecast$observed, exp_act_forecast$oracle_value)
})

#' Score model output predictions
#'
#' Scores model outputs with a single `output_type` against observed data.
#'
#' @param model_out_tbl Model output tibble with predictions
#' @param oracle_output Predictions that would have been generated by an oracle
#' model that knew the observed target data values in advance
#' @param metrics Character vector of scoring metrics to compute. If `NULL`
#' (the default), appropriate metrics are chosen automatically. See details
#' for more.
#' @param relative_metrics Character vector of scoring metrics for which to
#' compute relative skill scores. The `relative_metrics` should be a subset of
#' `metrics` and should only include proper scores. Interval coverage metrics
#' (e.g., `"interval_coverage_90"`) and `"bias"` are not supported as relative
#' metrics: interval coverage targets a nominal level rather than a "lower is
#' better" direction, and bias is a signed quantity for which a geometric-mean
#' pairwise ratio has no clean interpretation. If `NULL` (the default), no
#' relative metrics will be computed. Relative metrics are only computed if
#' `summarize = TRUE`, and require that `"model_id"` is included in `by`.
#' Relative skill is computed within each comparison group (the `by` columns
#' other than `"model_id"`). If a comparison group contains only one model,
#' that group's relative-skill columns are filled with `1` (a model has trivial
#' skill `1` relative to itself) rather than erroring.
#' @param baseline String with the name of a model to use as a baseline for
#' relative skill scores. If a baseline is given, then a scaled relative skill
#' with respect to the baseline will be returned. By default (`NULL`), relative
#' skill will not be scaled with respect to a baseline model. If the baseline is
#' absent from a comparison group, that group's relative and scaled relative
#' skill are reported as `NA` (with a warning) while its absolute scores are
#' still returned. This `NA` case only arises when scoring is disaggregated:
#' with `by = "model_id"` alone there is a single comparison group containing
#' every model, so relative skill is always defined and is never `NA`. A
#' baseline absent from the data entirely is an error.
#' @param summarize Boolean indicator of whether summaries of forecast scores
#' should be computed. Defaults to `TRUE`.
#' @param by Character vector naming columns to summarize by. For example,
#' specifying `by = "model_id"` (the default) will compute average scores for
#' each model.
#' @param output_type_id_order For ordinal variables in pmf format, this is a
#' vector of levels for pmf forecasts, in increasing order of the levels. The
#' order of the values for the output_type_id can be found by referencing the
#' hub's tasks.json configuration file. For all output types other than pmf,
#' this is ignored.
#' @param compound_taskid_set When `NULL` (the default), sample forecasts are
#' scored marginally (each modeling task scored independently). When a character
#' vector of task ID column names is provided, it sets the compound grouping:
#' the specified columns stay constant within each sample draw, while the
#' remaining task ID dimensions vary within a draw and are scored jointly
#' (using the energy score). Only applicable when
#' `output_type == "sample"`. The value of `compound_taskid_set` can be found
#' by referencing the hub's `tasks.json` configuration file.
#' @param transform A function to apply as a scale transformation to both
#'   predictions and observations before scoring. Common choices include
#'   \code{\link[scoringutils]{log_shift}} (recommended for log transformation
#'   as it handles zeros via an offset parameter), \code{sqrt}, or
#'   \code{log1p}. Avoid using \code{log} directly if data may contain zeros.
#'   If `NULL` (the default), no transformation is applied. Only supported for
#'   quantile, mean, and median output types.
#' @param transform_append Logical. If `FALSE` (the default), scores are
#'   computed only on the transformed scale. If `TRUE`, scores are computed on
#'   both original and transformed scales, with a `scale` column distinguishing
#'   them. Ignored if `transform = NULL`.
#' @param transform_label A character string label for the transformation
#'   (e.g., "log"). If `NULL` (the default), the label is auto-generated from
#'   the function name (e.g., "log_shift" for `scoringutils::log_shift`).
#'   Required when using an anonymous transform function. Ignored if
#'   `transform = NULL`. Note: the label only appears in output when
#'   `transform_append = TRUE`, where it distinguishes transformed rows
#'   (labeled with this value) from original rows (labeled "natural") in the
#'   `scale` column.
#' @param ... Additional arguments passed to the `transform` function. For
#'   example, allows use of the `offset` and `base` arguments of
#'   [scoringutils::log_shift()]. Ignored if `transform = NULL`.
#'
#' @details
#' See the hubverse documentation for the expected format of the
#' [oracle output data](https://docs.hubverse.io/en/latest/user-guide/target-data.html#oracle-output).
#'
#' Default metrics are provided by the `scoringutils` package. You can select
#' metrics by passing in a character vector of metric names to the `metrics`
#' argument.
#'
#' The following metrics can be selected (all are used by default) for the
#' different `output_type`s:
#'
#' **Quantile forecasts:** (`output_type == "quantile"`)
#' `r ex <- c("interval_coverage_50", "interval_coverage_90")`
#' `r .metrics_list(scoringutils::example_quantile, exclude = ex)`
#' - "interval_coverage_XX": interval coverage at the "XX" level. For example,
#' "interval_coverage_95" is the 95% interval coverage rate, which would be calculated
#' based on quantiles at the probability levels 0.025 and 0.975.
#'
#' See [scoringutils::get_metrics.forecast_quantile] for details.
#'
#' **Nominal forecasts:** (`output_type == "pmf"` and `output_type_id_order` is `NULL`)
#'
#' `r .metrics_list(scoringutils::example_nominal)`
#'
#' See [scoringutils::get_metrics.forecast_nominal] for details.
#'
#' **Ordinal forecasts:** (`output_type == "pmf"` and `output_type_id_order` is a vector)
#'
#' `r .metrics_list(scoringutils::example_ordinal)`
#'
#' See [scoringutils::get_metrics.forecast_ordinal] for details.
#'
#' **Median forecasts:** (`output_type == "median"`)
#'
#' - ae_point: absolute error of the point forecast (recommended for the median, see Gneiting (2011))
#'
#' See [scoringutils::get_metrics.forecast_point] for details.
#'
#' **Mean forecasts:** (`output_type == "mean"`)
#' - `se_point`: squared error of the point forecast (recommended for the mean, see Gneiting (2011))
#'
#' **Sample forecasts (marginal):** (`output_type == "sample"`, `compound_taskid_set = NULL`)
#'
#' `r .metrics_list(scoringutils::example_sample_continuous)`
#'
#' Note: `log_score` uses kernel density estimation, which may not be
#' appropriate for integer-valued forecasts. `scoringutils` will warn when
#' this is detected.
#'
#' See [scoringutils::get_metrics.forecast_sample] for details.
#'
#' **Sample forecasts (compound):** (`output_type == "sample"`, `compound_taskid_set` provided)
#'
#' `r .metrics_list(scoringutils::example_multivariate_sample)`
#'
#' See [scoringutils::get_metrics.forecast_multivariate_sample] for details.
#' The output includes a `.mv_group_id` column assigned by `scoringutils` to
#' identify the multivariate groups used for scoring (equivalent to the
#' `compound_idx` concept in the
#' [hubverse sample output type documentation
#' ](https://docs.hubverse.io/en/latest/user-guide/sample-output-type.html)).
#' Correct scoring depends on providing the right `compound_taskid_set` from
#' the hub's `tasks.json` configuration. If the specified grouping does not
#' match the actual dependence structure of the submitted samples (e.g.,
#' because some models submitted coarser samples than configured),
#' `.mv_group_id` may not correspond to the original sample draws as
#' indicated by their `output_type_id` values.
#'
#' See [scoringutils::add_relative_skill] for details on relative skill scores.
#'
#' @examplesIf requireNamespace("hubExamples", quietly = TRUE)
#' # compute WIS and interval coverage rates at 80% and 90% levels based on
#' # quantile forecasts, summarized by the mean score for each model
#' quantile_scores <- score_model_out(
#'   model_out_tbl = hubExamples::forecast_outputs |>
#'     dplyr::filter(.data[["output_type"]] == "quantile"),
#'   oracle_output = hubExamples::forecast_oracle_output,
#'   metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
#'   relative_metrics = "wis",
#'   by = "model_id"
#' )
#' quantile_scores
#'
#' # compute log scores based on pmf predictions for categorical targets,
#' # summarized by the mean score for each combination of model and location.
#' # Note: if the model_out_tbl had forecasts for multiple targets using a
#' # pmf output_type with different bins, it would be necessary to score the
#' # predictions for those targets separately.
#' pmf_scores <- score_model_out(
#'   model_out_tbl = hubExamples::forecast_outputs |>
#'     dplyr::filter(.data[["output_type"]] == "pmf"),
#'   oracle_output = hubExamples::forecast_oracle_output,
#'   metrics = c("log_score", "rps"),
#'   by = c("model_id", "location", "horizon"),
#'   output_type_id_order = c("low", "moderate", "high", "very high")
#' )
#' head(pmf_scores)
#'
#' # Score sample forecasts marginally (each modeling task scored independently).
#' # Note: this data has compound structure (samples span horizons), but marginal
#' # scoring is still valid -- it evaluates each horizon independently.
#' sample_scores <- score_model_out(
#'   model_out_tbl = hubExamples::forecast_outputs |>
#'     dplyr::filter(.data[["output_type"]] == "sample"),
#'   oracle_output = hubExamples::forecast_oracle_output,
#'   metrics = "crps",
#'   by = "model_id"
#' )
#' sample_scores
#'
#' # Score compound sample forecasts jointly using the energy score.
#' # compound_taskid_set specifies which task IDs stay constant within
#' # a sample group -- here, each sample draw spans all horizons for a
#' # given reference_date and location (i.e., a trajectory over time).
#' compound_scores <- score_model_out(
#'   model_out_tbl = hubExamples::forecast_outputs |>
#'     dplyr::filter(.data[["output_type"]] == "sample"),
#'   oracle_output = hubExamples::forecast_oracle_output,
#'   compound_taskid_set = c("reference_date", "location"),
#'   by = "model_id"
#' )
#' compound_scores
#'
#' @return A tibble of scores, inheriting from scoringutils' `scores` class so
#' that downstream scoringutils helpers (e.g. [scoringutils::get_metrics()])
#' continue to work. The tibble has a `metrics` attribute holding the names of
#' the scoring rules that were applied.
#'
#' @references
#' Gneiting, Tilmann. 2011. "Making and Evaluating Point Forecasts." Journal of the
#' American Statistical Association 106 (494): 746–62. \doi{10.1198/jasa.2011.r10138}.
#'
#' @export
score_model_out <- function(
  model_out_tbl,
  oracle_output,
  metrics = NULL,
  relative_metrics = NULL,
  baseline = NULL,
  summarize = TRUE,
  by = "model_id",
  output_type_id_order = NULL,
  compound_taskid_set = NULL,
  transform = NULL,
  transform_append = FALSE,
  transform_label = NULL,
  ...
) {
  # check that model_out_tbl has a single output_type that is supported by this package
  # also, retrieve that output_type
  output_type <- validate_output_type(model_out_tbl)

  if (!is.null(compound_taskid_set) && output_type != "sample") {
    cli::cli_abort(
      "{.arg compound_taskid_set} is only applicable to sample output types,
      but {.arg model_out_tbl} has output type {.val {output_type}}."
    )
  }

  # Validate transformation parameters and capture expression for label inference
  transform_expr <- rlang::enexpr(transform)
  validate_transform(transform, transform_label, output_type)
  if (summarize) {
    # Note: The call to scoringutils::add_relative_skill below performs validation of `baseline`
    validate_relative_metrics(relative_metrics, metrics, by)
  }

  # assemble data for scoringutils
  su_data <- switch(
    output_type,
    quantile = transform_quantile_model_out(model_out_tbl, oracle_output),
    pmf = transform_pmf_model_out(
      model_out_tbl,
      oracle_output,
      output_type_id_order
    ),
    mean = transform_point_model_out(model_out_tbl, oracle_output, output_type),
    median = transform_point_model_out(
      model_out_tbl,
      oracle_output,
      output_type
    ),
    sample = transform_sample_model_out(
      model_out_tbl,
      oracle_output,
      compound_taskid_set
    ),
    NULL # default, should not happen because of the validation above
  )

  # Apply scale transformation if requested
  if (!is.null(transform)) {
    # Use the captured expression for label inference (e.g., "log_shift")
    label <- get_transform_label(transform_expr, transform_label)
    su_data <- scoringutils::transform_forecasts(
      su_data,
      fun = transform,
      append = transform_append,
      label = label,
      ...
    )
    validate_transform_values(su_data)
  }

  # get/validate the scoring metrics
  metrics <- get_metrics(
    forecast = su_data,
    output_type = output_type,
    select = metrics
  )

  # compute scores
  scores <- scoringutils::score(su_data, metrics)

  # switch back to hubverse naming conventions for model name
  scores <- dplyr::rename(scores, model_id = "model")

  # if requested, summarize scores, including computation of relative metrics
  if (summarize) {
    # Relative skill is a per-group quantity (one value per model per `by`
    # group, not per scored forecast unit). It is added here on the per-row
    # scores as a column that is constant within each (model_id, `by`) group,
    # then collapsed by summarize_scores() below to one row per group. Pairwise
    # comparison needs the per-row scores, so this must run before summarising.
    scores <- add_relative_skill_metrics(scores, relative_metrics, by, baseline)
    # BEGIN workaround mirror for https://github.com/epiforecasts/scoringutils/pull/1180
    # Mirror the upstream check (merged but not yet on CRAN) so the
    # user-facing behaviour of score_model_out() does not shift once we
    # bump the scoringutils minimum version and delete this block. Wording
    # and condition match the upstream implementation exactly.
    metric_cols <- intersect(colnames(scores), attr(scores, "metrics"))
    if (length(metric_cols) == 0) {
      cli::cli_abort(
        c(
          `!` = "No score columns to summarise.",
          i = "The {.cls scores} object has no columns matching its
               {.code metrics} attribute. This usually means every metric
               passed to {.fn score} failed (e.g. warned and returned no
               values)."
        )
      )
    }
    # END workaround mirror for scoringutils#1180

    # When transform_append = TRUE, scoringutils emits both natural- and
    # transformed-scale rows distinguished by a `scale` column. Include
    # "scale" in `by` for the summary so the two scales are reported
    # separately rather than collapsed into a meaningless average. The
    # user-facing `by` argument is not modified outside this block.
    summary_by <- if (isTRUE(transform_append)) union(by, "scale") else by
    scores <- scoringutils::summarize_scores(scores = scores, by = summary_by)
  }

  # Convert to tibble for friendlier user-facing behaviour, but keep the
  # `scores` class on top so scoringutils helpers (e.g. get_metrics) still
  # dispatch. The `metrics` attribute is preserved by as_tibble().
  scores <- tibble::as_tibble(scores)
  class(scores) <- c("scores", class(scores))
  scores
}


#' Add relative-skill columns to a per-row scores object
#'
#' Validates that any requested `baseline` is present in the data (a baseline
#' absent entirely is a user error), then delegates to
#' [add_relative_skill_by_group()], which computes relative skill within each
#' comparison group and gracefully handles groups scoringutils cannot compare
#' (single-model groups, or groups missing the baseline) instead of aborting.
#'
#' @noRd
add_relative_skill_metrics <- function(scores, relative_metrics, by, baseline) {
  if (length(relative_metrics) == 0L) {
    return(scores)
  }

  models <- unique(scores[["model_id"]])

  # A baseline must exist somewhere in the data. (A baseline missing from only
  # *some* comparison groups is handled gracefully by
  # add_relative_skill_by_group(); one absent entirely is a typo-style user
  # error and aborts here.)
  if (!is.null(baseline) && !baseline %in% models) {
    cli::cli_abort(c(
      "Requested {.arg baseline} {.val {baseline}} is not present in the data.",
      "i" = "Available models: {.val {models}}."
    ))
  }

  add_relative_skill_by_group(scores, relative_metrics, by, baseline)
}


#' Column names added by relative-skill scoring
#'
#' For each requested metric, scoringutils adds `<metric>_relative_skill`, plus
#' `<metric>_scaled_relative_skill` when a `baseline` is supplied. Returns those
#' names in the order the columns are produced.
#'
#' @noRd
relative_skill_col_names <- function(relative_metrics, baseline) {
  suffixes <- if (is.null(baseline)) {
    "_relative_skill"
  } else {
    c("_relative_skill", "_scaled_relative_skill")
  }
  purrr::map(relative_metrics, \(metric) paste0(metric, suffixes)) |>
    purrr::list_c()
}


#' Add relative-skill columns, classifying groups by how skill is produced
#'
#' Relative skill is computed within each comparison group: one combination of
#' the disaggregation columns (`by` without `"model_id"`), or a single group
#' spanning the whole data when there is no disaggregation. A group scoringutils
#' cannot score would otherwise abort the whole run (e.g. "Baseline comparator
#' `<name>` missing."), so each group is classified and handled instead.
#' Predictions in every group are still scoreable, so absolute scores stay
#' intact; only relative skill needs special handling.
#'
#' Each comparison group is one of:
#' * **computable** (>= 2 models, and the baseline present if one is requested):
#'   scored by scoringutils.
#' * **trivial** (a single model, with no baseline requested or where that lone
#'   model is the baseline): relative skill is `1` (a model has skill `1`
#'   relative to itself). This subsumes the single-model-everywhere case.
#' * **skipped** (a baseline was requested but is absent from the group, e.g. a
#'   baseline missing some of the `reference_date`s being scored when
#'   disaggregating by `reference_date`): relative skill is `NA`, with one
#'   warning naming the affected group levels. See #135.
#'
#' The caller ([add_relative_skill_metrics()]) has already checked that any
#' `baseline` is present somewhere in the data.
#'
#' @return `scores` with relative-skill columns added.
#' @noRd
add_relative_skill_by_group <- function(
  scores,
  relative_metrics,
  by,
  baseline
) {
  group_cols <- by[by != "model_id"]
  orig_metrics <- attr(scores, "metrics")
  skill_cols <- relative_skill_col_names(relative_metrics, baseline)

  # Classify each comparison group by how its relative skill must be produced
  # (see the function docs): computable groups are scored, trivial single-model
  # groups get 1, and groups missing a requested baseline are skipped to NA.
  # With no disaggregation `group_cols` is empty and this is a single group
  # spanning the whole data: that case (and single-model data) runs through
  # this same path rather than a fast early return, so the two cannot drift
  # apart, which is worth more than the small summarise/join overhead.
  group_status <- scores |>
    dplyr::group_by(dplyr::across(dplyr::all_of(group_cols))) |>
    dplyr::summarise(
      has_baseline = is.null(baseline) || baseline %in% .data[["model_id"]],
      n_models = dplyr::n_distinct(.data[["model_id"]]),
      .groups = "drop"
    )
  has_baseline <- group_status[["has_baseline"]]
  n_models <- group_status[["n_models"]]
  computable_groups <- group_status[
    has_baseline & n_models >= 2L,
    group_cols,
    drop = FALSE
  ]
  trivial_groups <- group_status[
    has_baseline & n_models < 2L,
    group_cols,
    drop = FALSE
  ]
  skipped_groups <- group_status[!has_baseline, group_cols, drop = FALSE]

  # Assemble one (model_id, group) -> relative-skill lookup, then broadcast it
  # onto the full scores. Skipped groups are absent from the lookup, so they
  # find no match in the join and become NA by construction.
  lookup <- dplyr::bind_rows(
    compute_skills(
      scores,
      computable_groups,
      group_cols,
      skill_cols,
      relative_metrics,
      baseline
    ),
    assign_trivial_skills(scores, trivial_groups, group_cols, skill_cols)
  )
  scores <- dplyr::left_join(scores, lookup, by = c("model_id", group_cols))
  attr(scores, "metrics") <- c(orig_metrics, skill_cols)

  if (nrow(skipped_groups) > 0L) {
    warn_relative_skill_skipped(skipped_groups, group_cols, baseline)
  }
  scores
}


#' Subset scores to the rows whose group is in `groups`
#'
#' A `semi_join` on `group_cols`. When `group_cols` is empty (no disaggregation)
#' there is a single group spanning all of `scores`, so there is nothing to
#' subset by and every row is returned unchanged; handling that explicitly also
#' avoids dplyr's deprecated empty-key cross join.
#'
#' @noRd
subset_scores_to_groups <- function(scores, groups, group_cols) {
  if (length(group_cols) == 0L) {
    # No grouping columns: a single group spans all of `scores`, so every row
    # is already in it. There is nothing to subset by, so keep all rows.
    scores
  } else {
    dplyr::semi_join(scores, groups, by = group_cols)
  }
}


#' Per-(model_id, group) relative skill for groups scoringutils can score
#'
#' Scores the computable subset with [scoringutils::add_relative_skill()] (once
#' per metric) and reduces it to one row per `(model_id, group)`, since relative
#' skill is constant within a group. Returns `NULL` when there are no computable
#' groups, so [dplyr::bind_rows()] drops it.
#'
#' @param scores The full per-row scores object.
#' @param computable_groups One row per computable group, holding its
#'   `group_cols` values.
#' @param group_cols Character vector of disaggregation column names.
#' @param skill_cols Names of the relative-skill columns to return, as produced
#'   by [relative_skill_col_names()] (e.g. `"wis_relative_skill"`, plus
#'   `"wis_scaled_relative_skill"` when a baseline is given).
#' @param relative_metrics Character vector of metrics to score relative skill for.
#' @param baseline The requested baseline model, or `NULL`.
#' @noRd
compute_skills <- function(
  scores,
  computable_groups,
  group_cols,
  skill_cols,
  relative_metrics,
  baseline
) {
  if (nrow(computable_groups) == 0L) {
    return(NULL)
  }
  computable_scores <- subset_scores_to_groups(
    scores,
    computable_groups,
    group_cols
  )
  for (metric in relative_metrics) {
    computable_scores <- scoringutils::add_relative_skill(
      computable_scores,
      compare = "model_id",
      by = group_cols,
      metric = metric,
      baseline = baseline
    )
  }
  # The lookup we merge back onto the scored data needs only the join keys
  # (model_id, group_cols) and the skill columns. Scoring above used the per-row
  # metric columns, so we drop them and de-dup to one row per group here, after
  # add_relative_skill().
  computable_scores |>
    tibble::as_tibble() |>
    dplyr::select(dplyr::all_of(c("model_id", group_cols, skill_cols))) |>
    dplyr::distinct()
}


#' Per-(model_id, group) relative skill of 1 for single-model groups
#'
#' A group with one model has no comparator, so relative (and scaled relative)
#' skill is `1` by the single-model convention. Returns one row per `(model_id,
#' group)` with every relative-skill column set to `1`, or `NULL` when there are
#' no such groups.
#'
#' @param scores The full per-row scores object.
#' @param trivial_groups One row per single-model group, holding its
#'   `group_cols` values.
#' @param group_cols Character vector of disaggregation column names.
#' @param skill_cols Names of the relative-skill columns to set to `1`, as
#'   produced by [relative_skill_col_names()].
#' @noRd
assign_trivial_skills <- function(
  scores,
  trivial_groups,
  group_cols,
  skill_cols
) {
  if (nrow(trivial_groups) == 0L) {
    return(NULL)
  }
  # The same merge-back lookup (join keys + skill columns, one row per group),
  # but nothing is scored here, so unlike compute_skills we drop the metric
  # columns and de-dup to the keys up front, then assign 1 to the skill columns.
  trivial_scores <- subset_scores_to_groups(
    scores,
    trivial_groups,
    group_cols
  ) |>
    tibble::as_tibble() |>
    dplyr::distinct(dplyr::across(dplyr::all_of(c("model_id", group_cols))))
  for (col in skill_cols) {
    trivial_scores[[col]] <- 1
  }
  trivial_scores
}


#' Warn that relative skill could not be scaled to the baseline for some groups
#'
#' Only baseline-absent groups are skipped to `NA`; single-model groups are
#' filled with `1` and do not warn. A skipped group therefore always implies a
#' `baseline` was requested.
#'
#' @param skipped_groups Data frame of disaggregation-column values for the
#'   groups from which the baseline is absent.
#' @param group_cols Character vector of disaggregation column names.
#' @param baseline The requested baseline model.
#' @noRd
warn_relative_skill_skipped <- function(skipped_groups, group_cols, baseline) {
  labels <- apply(skipped_groups, 1L, function(row) {
    paste(group_cols, row, sep = " = ", collapse = ", ")
  })
  n <- length(labels)
  shown <- utils::head(labels, 5L)
  if (n > length(shown)) {
    shown <- c(shown, "...")
  }
  cli::cli_warn(c(
    "!" = "Relative skill could not be scaled to the baseline
           {.val {baseline}} for {n} disaggregation group{?s} with no baseline
           prediction; reporting NA for them.",
    "i" = "Affected levels of {.val {group_cols}}: {.val {shown}}."
  ))
}


#' Get scoring metrics
#'
#' @param forecast A scoringutils `forecast` object (see
#' [scoringutils' general information on creating a forecast object][scoringutils::as_forecast_doc_template()]
#' for details).
#' @inheritParams score_model_out
#'
#' @return a list of metric functions as required by scoringutils::score()
#'
#' @noRd
get_metrics <- function(forecast, output_type, select = NULL) {
  forecast_type <- class(forecast)[1]

  # process quantile metrics separately to allow better selection of interval
  # coverage metrics
  if (forecast_type == "forecast_quantile") {
    # split into metrics for interval coverage and others
    interval_metric_inds <- is_interval_coverage_metric(select)
    interval_metrics <- select[interval_metric_inds]
    other_metrics <- select[!interval_metric_inds]

    other_metric_fns <- scoringutils::get_metrics(
      forecast,
      select = other_metrics
    )

    # assemble interval coverage functions
    interval_metric_fns <- lapply(
      interval_metrics,
      function(metric) {
        level_str <- substr(metric, 19, nchar(metric))
        level <- suppressWarnings(as.numeric(level_str))
        if (is.na(level) || level <= 0 || level >= 100) {
          cli::cli_abort(c(
            "Invalid interval coverage level: {level_str}",
            "i" = "must be a number between 0 and 100 (exclusive)"
          ))
        }
        purrr::partial(scoringutils::interval_coverage, interval_range = level)
      }
    )
    names(interval_metric_fns) <- interval_metrics

    metric_fns <- c(other_metric_fns, interval_metric_fns)
    return(metric_fns)
  }

  # leave validation of user selection to scoringutils
  metric_fns <- scoringutils::get_metrics(forecast, select = select)
  if (output_type == "mean") {
    metric_fns <- scoringutils::select_metrics(metric_fns, "se_point")
  } else if (output_type == "median") {
    metric_fns <- scoringutils::select_metrics(metric_fns, "ae_point")
  }

  metric_fns
}


#' Determine whether metric is an interval coverage metric
#'
#' @param metrics Character vector of metric names
#' @return Logical vector indicating whether each metric is an interval coverage metric
#' @noRd
is_interval_coverage_metric <- function(metric) {
  grepl(pattern = "^interval_coverage_", metric)
}

test_that("score_model_out succeeds with valid inputs: quantile output_type, relative wis and ae, no baseline", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # Wilcoxon test warns about ties/zeroes on small samples; expected, not a bug
  act_scores <- suppress_wilcox_ties_warnings(score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "quantile"),
    oracle_output = forecast_oracle_output,
    metrics = c(
      "ae_median",
      "wis",
      "interval_coverage_80",
      "interval_coverage_90"
    ),
    relative_metrics = c("ae_median", "wis"),
    by = c("model_id", "location")
  ))

  exp_scores <- read.csv(test_path("testdata", "exp_pairwise_scores.csv")) |>
    dplyr::mutate(location = as.character(location)) |>
    dplyr::select(-ae_median_scaled_relative_skill, -wis_scaled_relative_skill)

  expect_equal(act_scores, exp_scores, ignore_attr = TRUE)
})


test_that("score_model_out succeeds with valid inputs: quantile output_type, relative wis and ae, Flusight-baseline", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # Wilcoxon test warns about ties/zeroes on small samples; expected, not a bug
  act_scores <- suppress_wilcox_ties_warnings(score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "quantile"),
    oracle_output = forecast_oracle_output,
    metrics = c(
      "ae_median",
      "wis",
      "interval_coverage_80",
      "interval_coverage_90"
    ),
    relative_metrics = c("ae_median", "wis"),
    baseline = "Flusight-baseline",
    by = c("model_id", "location")
  ))

  exp_scores <- read.csv(test_path("testdata", "exp_pairwise_scores.csv")) |>
    dplyr::mutate(location = as.character(location))

  expect_equal(act_scores, exp_scores, ignore_attr = TRUE)
})


test_that("score_model_out errors when invalid relative metrics are requested", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # not allowed to compute relative skill for interval coverage
  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
      relative_metrics = c("interval_coverage_90", "wis"),
    ),
    regexp = "Interval coverage metrics are not supported for relative skill scores."
  )

  # not allowed to compute relative skill for bias: hubEvals errors at the
  # boundary, before scoringutils fails with its "same sign" message.
  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")
  expect_error(
    score_model_out(
      model_out_tbl = quantile_out,
      oracle_output = forecast_oracle_output,
      metrics = c("wis", "bias"),
      relative_metrics = "bias"
    ),
    regexp = "bias.*not supported for relative skill scores"
  )

  # relative_metrics must be a subset of metrics
  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
      relative_metrics = c("ae_median", "wis"),
    ),
    regexp = "Relative metrics must be a subset of the metrics."
  )

  # can't ask for relative metrics without breaking down by model_id
  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
      relative_metrics = "wis",
      by = "location"
    ),
    regexp = "Relative metrics require 'model_id' to be included in `by`."
  )
})


test_that("score_model_out handles single-model input gracefully for relative metrics", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  single <- forecast_outputs |>
    dplyr::filter(
      .data[["output_type"]] == "quantile",
      .data[["model_id"]] == "Flusight-baseline"
    )

  # No baseline: just fills in the relative-skill column with 1
  scores_no_baseline <- score_model_out(
    model_out_tbl = single,
    oracle_output = forecast_oracle_output,
    metrics = c("wis", "ae_median"),
    relative_metrics = c("wis", "ae_median")
  )
  expect_equal(nrow(scores_no_baseline), 1L)
  expect_equal(scores_no_baseline$wis_relative_skill, 1)
  expect_equal(scores_no_baseline$ae_median_relative_skill, 1)
  expect_false("wis_scaled_relative_skill" %in% names(scores_no_baseline))
  expect_true(all(
    c("wis_relative_skill", "ae_median_relative_skill") %in%
      attr(scores_no_baseline, "metrics")
  ))

  # Baseline matching the lone model: both relative_skill and scaled_relative_skill = 1
  scores_baseline <- score_model_out(
    model_out_tbl = single,
    oracle_output = forecast_oracle_output,
    metrics = "wis",
    relative_metrics = "wis",
    baseline = "Flusight-baseline"
  )
  expect_equal(scores_baseline$wis_relative_skill, 1)
  expect_equal(scores_baseline$wis_scaled_relative_skill, 1)
  expect_true(all(
    c("wis_relative_skill", "wis_scaled_relative_skill") %in%
      attr(scores_baseline, "metrics")
  ))

  # Baseline that is not the lone model: errors
  expect_error(
    score_model_out(
      model_out_tbl = single,
      oracle_output = forecast_oracle_output,
      metrics = "wis",
      relative_metrics = "wis",
      baseline = "MOBS-GLEAM_FLUH"
    ),
    regexp = "baseline.*MOBS-GLEAM_FLUH.*not present"
  )
})


test_that("score_model_out reports NA relative skill and warns when a baseline is missing from a group", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Drop the baseline's predictions for one location entirely, so the location
  # 25 comparison group (defined by `by`) contains no baseline.
  is_dropped <- quantile_out[["model_id"]] == "Flusight-baseline" &
    quantile_out[["location"]] == "25"
  incomplete <- quantile_out[!is_dropped, ]

  # Absolute scores computed without relative metrics, as the reference point
  # for "absolute scores and counts unchanged".
  base_scores <- suppress_wilcox_ties_warnings(score_model_out(
    model_out_tbl = incomplete,
    oracle_output = forecast_oracle_output,
    metrics = c("wis", "ae_median"),
    by = c("model_id", "location")
  ))

  expect_warning(
    rel_scores <- score_model_out(
      model_out_tbl = incomplete,
      oracle_output = forecast_oracle_output,
      metrics = c("wis", "ae_median"),
      relative_metrics = "wis",
      baseline = "Flusight-baseline",
      by = c("model_id", "location")
    ),
    regexp = "Flusight-baseline.*location"
  )

  # Relative skill is NA exactly for the baseline-absent group (location 25)
  # and finite for the group the baseline covers (location 48).
  loc25 <- rel_scores[rel_scores[["location"]] == "25", ]
  loc48 <- rel_scores[rel_scores[["location"]] == "48", ]
  expect_true(all(is.na(loc25[["wis_relative_skill"]])))
  expect_true(all(is.na(loc25[["wis_scaled_relative_skill"]])))
  expect_true(all(is.finite(loc48[["wis_relative_skill"]])))
  expect_true(all(is.finite(loc48[["wis_scaled_relative_skill"]])))

  # Absolute scores and row counts are unchanged by requesting relative metrics.
  expect_equal(nrow(rel_scores), nrow(base_scores))
  key <- c("model_id", "location")
  joined <- dplyr::left_join(
    base_scores[, c(key, "wis", "ae_median")],
    rel_scores[, c(key, "wis", "ae_median")],
    by = key,
    suffix = c("_base", "_rel")
  )
  expect_equal(joined[["wis_base"]], joined[["wis_rel"]])
  expect_equal(joined[["ae_median_base"]], joined[["ae_median_rel"]])
})


test_that("score_model_out still computes relative skill when a baseline has partial (within-group) coverage", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Drop only some of the baseline's location 25 predictions (one reference
  # date), so the baseline is still present in the location 25 group. Within a
  # group scoringutils handles partial overlap, so this must still compute.
  is_dropped <- quantile_out[["model_id"]] == "Flusight-baseline" &
    quantile_out[["location"]] == "25" &
    quantile_out[["reference_date"]] == "2022-11-19"
  partial <- quantile_out[!is_dropped, ]

  # No baseline-coverage warning, and relative skill computed for every group.
  # The Wilcoxon ties warning on this small sample is expected and suppressed.
  scores <- suppress_wilcox_ties_warnings(
    expect_no_warning(
      score_model_out(
        model_out_tbl = partial,
        oracle_output = forecast_oracle_output,
        metrics = "wis",
        relative_metrics = "wis",
        baseline = "Flusight-baseline",
        by = c("model_id", "location")
      ),
      message = "could not be scaled"
    )
  )
  expect_true(all(is.finite(scores[["wis_relative_skill"]])))
})


test_that("score_model_out errors when the baseline is absent from the data entirely", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # A baseline name that appears nowhere in the data is a user error, not an
  # incomplete-coverage case, and gets a clear hubEvals message rather than
  # scoringutils' checkmate assertion.
  expect_error(
    suppress_wilcox_ties_warnings(score_model_out(
      model_out_tbl = quantile_out,
      oracle_output = forecast_oracle_output,
      metrics = "wis",
      relative_metrics = "wis",
      baseline = "not-a-real-model",
      by = c("model_id", "location")
    )),
    regexp = "baseline.*not-a-real-model.*not present"
  )
})


test_that("score_model_out computes relative skill without a coverage warning when the baseline covers every group", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Wilcoxon test warns about ties/zeroes on small samples; expected and
  # suppressed. Assert only that the baseline-coverage warning is absent.
  suppress_wilcox_ties_warnings(
    expect_no_warning(
      score_model_out(
        model_out_tbl = quantile_out,
        oracle_output = forecast_oracle_output,
        metrics = c("wis", "ae_median"),
        relative_metrics = "wis",
        baseline = "Flusight-baseline",
        by = c("model_id", "location")
      ),
      message = "could not be scaled"
    )
  )
})


test_that("score_model_out computes relative skill with no disaggregation when the baseline has partial coverage", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # With `by = "model_id"` there is a single comparison group spanning the whole
  # data, so a baseline that is missing an entire location is still present in
  # that one group. scoringutils compares each pair on its overlapping forecast
  # units, so relative skill is computed (no NA, no coverage warning).
  is_dropped <- quantile_out[["model_id"]] == "Flusight-baseline" &
    quantile_out[["location"]] == "25"
  partial <- quantile_out[!is_dropped, ]

  # The Wilcoxon ties warning is expected and suppressed; assert only that the
  # baseline-coverage warning is absent.
  scores <- suppress_wilcox_ties_warnings(
    expect_no_warning(
      score_model_out(
        model_out_tbl = partial,
        oracle_output = forecast_oracle_output,
        metrics = "wis",
        relative_metrics = "wis",
        baseline = "Flusight-baseline",
        by = "model_id"
      ),
      message = "could not be scaled"
    )
  )

  # One summarised row per model, all relative skill finite.
  expect_equal(nrow(scores), 3L)
  expect_true(all(is.finite(scores[["wis_relative_skill"]])))
  expect_true(all(is.finite(scores[["wis_scaled_relative_skill"]])))
})


test_that("score_model_out fills relative skill with 1 for single-model disaggregation groups", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Make location 25 a single-model group (only MOBS-GLEAM_FLUH); location 48
  # keeps all models. A lone-model group has no comparator, so -- consistent
  # with the global single-model case -- its relative skill is 1, not NA, and
  # it does not warn.
  drop_others <- quantile_out[["location"]] == "25" &
    quantile_out[["model_id"]] != "MOBS-GLEAM_FLUH"
  one_model_loc <- quantile_out[!drop_others, ]

  scores <- suppress_wilcox_ties_warnings(
    expect_no_warning(
      score_model_out(
        model_out_tbl = one_model_loc,
        oracle_output = forecast_oracle_output,
        metrics = "wis",
        relative_metrics = "wis",
        by = c("model_id", "location")
      ),
      message = "could not be scaled"
    )
  )

  loc25 <- scores[scores[["location"]] == "25", ]
  loc48 <- scores[scores[["location"]] == "48", ]
  expect_equal(loc25[["wis_relative_skill"]], 1)
  expect_true(all(is.finite(loc48[["wis_relative_skill"]])))
})


test_that("score_model_out fills 1 for a single-model group made up of the baseline", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Location 25 contains only the baseline model. The baseline is present in
  # that group (so it is not skipped to NA) but has no comparator, so relative
  # and scaled relative skill are both 1, with no warning.
  drop_others <- quantile_out[["location"]] == "25" &
    quantile_out[["model_id"]] != "Flusight-baseline"
  baseline_only_loc <- quantile_out[!drop_others, ]

  scores <- suppress_wilcox_ties_warnings(
    expect_no_warning(
      score_model_out(
        model_out_tbl = baseline_only_loc,
        oracle_output = forecast_oracle_output,
        metrics = "wis",
        relative_metrics = "wis",
        baseline = "Flusight-baseline",
        by = c("model_id", "location")
      ),
      message = "could not be scaled"
    )
  )

  loc25 <- scores[scores[["location"]] == "25", ]
  expect_equal(loc25[["wis_relative_skill"]], 1)
  expect_equal(loc25[["wis_scaled_relative_skill"]], 1)
})

test_that("score_model_out succeeds with valid inputs: mean output_type, default metrics, by all", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "mean"),
    oracle_output = forecast_oracle_output,
    by = c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  )

  exp_scores <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "mean") |>
    dplyr::left_join(
      forecast_oracle_output |>
        dplyr::filter(.data[["output_type"]] == "mean"),
      by = c("location", "target_end_date", "target")
    ) |>
    dplyr::mutate(
      se = (.data[["value"]] - .data[["oracle_value"]])^2
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c(
        "model_id",
        "location",
        "reference_date",
        "horizon",
        "target_end_date",
        "target"
      )
    ))) |>
    dplyr::summarize(
      se_point = mean(.data[["se"]]),
      .groups = "drop"
    )

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  expect_equal(merged_scores$se_point.x, merged_scores$se_point.y)
})


test_that("score_model_out succeeds with valid inputs: mean output_type, default metrics, summarize FALSE", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "mean"),
    oracle_output = forecast_oracle_output,
    summarize = FALSE
  )

  exp_scores <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "mean") |>
    dplyr::left_join(
      forecast_oracle_output |>
        dplyr::filter(.data[["output_type"]] == "mean"),
      by = c("location", "target_end_date", "target")
    ) |>
    dplyr::mutate(
      se = (.data[["value"]] - .data[["oracle_value"]])^2
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c(
        "model_id",
        "reference_date",
        "target",
        "horizon",
        "location",
        "target_end_date"
      )
    ))) |>
    dplyr::summarize(
      se_point = mean(.data[["se"]]),
      .groups = "drop"
    )

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  expect_equal(merged_scores$se_point.x, merged_scores$se_point.y)
})


test_that("score_model_out succeeds with valid inputs: mean output_type, character metrics, custom by", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "mean"),
    oracle_output = forecast_oracle_output,
    metrics = c("ae_point", "se_point"),
    by = c("model_id", "location")
  )

  exp_scores <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "mean") |>
    dplyr::left_join(
      forecast_oracle_output |>
        dplyr::filter(.data[["output_type"]] == "mean"),
      by = c("location", "target_end_date", "target")
    ) |>
    dplyr::mutate(
      ae = abs(.data[["value"]] - .data[["oracle_value"]]),
      se = (.data[["value"]] - .data[["oracle_value"]])^2
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c("model_id", "location")
    ))) |>
    dplyr::summarize(
      se_point = mean(.data[["se"]]),
      .groups = "drop"
    )

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c("model_id", "location")
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  # `ae_point` is intentionally not returned for mean (see Gneiting 2011),
  # so only `se_point` is checked here even though it was also requested.
  expect_equal(merged_scores$se_point.x, merged_scores$se_point.y)
})


test_that("score_model_out succeeds with valid inputs: median output_type, default metrics, summarize FALSE", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "median"),
    oracle_output = forecast_oracle_output,
    summarize = FALSE
  )

  exp_scores <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "median") |>
    dplyr::left_join(
      forecast_oracle_output |>
        dplyr::filter(.data[["output_type"]] == "median"),
      by = c("location", "target_end_date", "target")
    ) |>
    dplyr::mutate(
      ae = abs(.data[["value"]] - .data[["oracle_value"]])
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c(
        "model_id",
        "reference_date",
        "target",
        "horizon",
        "location",
        "target_end_date"
      )
    ))) |>
    dplyr::summarize(
      ae_point = mean(.data[["ae"]]),
      .groups = "drop"
    )

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  expect_equal(merged_scores$ae_point.x, merged_scores$ae_point.y)
})


test_that("score_model_out succeeds with valid inputs: quantile output_type, wis and interval metrics, custom by", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "quantile"),
    oracle_output = forecast_oracle_output,
    metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
    by = c("model_id", "location")
  )

  exp_scores <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile") |>
    dplyr::left_join(
      forecast_oracle_output |>
        dplyr::filter(.data[["output_type"]] == "quantile") |>
        dplyr::select(-dplyr::all_of(c("output_type", "output_type_id"))),
      by = c("location", "target_end_date", "target")
    ) |>
    dplyr::mutate(
      output_type_id = as.numeric(.data[["output_type_id"]]),
      qs = ifelse(
        .data[["oracle_value"]] >= .data[["value"]],
        .data[["output_type_id"]] *
          (.data[["oracle_value"]] - .data[["value"]]),
        (1 - .data[["output_type_id"]]) *
          (.data[["value"]] - .data[["oracle_value"]])
      ),
      q_coverage_80_lower = ifelse(
        .data[["output_type_id"]] == 0.1,
        .data[["oracle_value"]] >= .data[["value"]],
        NA_real_
      ),
      q_coverage_80_upper = ifelse(
        .data[["output_type_id"]] == 0.9,
        .data[["oracle_value"]] <= .data[["value"]],
        NA_real_
      ),
      q_coverage_90_lower = ifelse(
        .data[["output_type_id"]] == 0.05,
        .data[["oracle_value"]] >= .data[["value"]],
        NA_real_
      ),
      q_coverage_90_upper = ifelse(
        .data[["output_type_id"]] == 0.95,
        .data[["oracle_value"]] <= .data[["value"]],
        NA_real_
      )
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c(
        "model_id",
        "location",
        "reference_date",
        "horizon",
        "target_end_date",
        "target"
      )
    ))) |>
    dplyr::summarize(
      wis = 2 * mean(.data[["qs"]]),
      interval_coverage_80 = (sum(
        .data[["q_coverage_80_lower"]],
        na.rm = TRUE
      ) ==
        1) *
        (sum(.data[["q_coverage_80_upper"]], na.rm = TRUE) == 1),
      interval_coverage_90 = (sum(
        .data[["q_coverage_90_lower"]],
        na.rm = TRUE
      ) ==
        1) *
        (sum(.data[["q_coverage_90_upper"]], na.rm = TRUE) == 1)
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c("model_id", "location")
    ))) |>
    dplyr::summarize(
      wis = mean(.data[["wis"]]),
      interval_coverage_80 = mean(
        .data[["interval_coverage_80"]],
        na.rm = TRUE
      ),
      interval_coverage_90 = mean(
        .data[["interval_coverage_90"]],
        na.rm = TRUE
      ),
      .groups = "drop"
    )

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c("model_id", "location")
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  expect_equal(merged_scores$wis.x, merged_scores$wis.y)
  expect_equal(
    merged_scores$interval_coverage_80.x,
    merged_scores$interval_coverage_80.y
  )
  expect_equal(
    merged_scores$interval_coverage_90.x,
    merged_scores$interval_coverage_90.y
  )
})


test_that("score_model_out succeeds with valid inputs: quantile output_type, wis/interval metrics, summarize FALSE", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "quantile"),
    oracle_output = forecast_oracle_output,
    metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
    summarize = FALSE
  )

  exp_scores <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile") |>
    dplyr::left_join(
      forecast_oracle_output |>
        dplyr::filter(.data[["output_type"]] == "quantile") |>
        dplyr::select(-dplyr::all_of(c("output_type", "output_type_id"))),
      by = c("location", "target_end_date", "target")
    ) |>
    dplyr::mutate(
      output_type_id = as.numeric(.data[["output_type_id"]]),
      qs = ifelse(
        .data[["oracle_value"]] >= .data[["value"]],
        .data[["output_type_id"]] *
          (.data[["oracle_value"]] - .data[["value"]]),
        (1 - .data[["output_type_id"]]) *
          (.data[["value"]] - .data[["oracle_value"]])
      ),
      q_coverage_80_lower = ifelse(
        .data[["output_type_id"]] == 0.1,
        .data[["oracle_value"]] >= .data[["value"]],
        NA_real_
      ),
      q_coverage_80_upper = ifelse(
        .data[["output_type_id"]] == 0.9,
        .data[["oracle_value"]] <= .data[["value"]],
        NA_real_
      ),
      q_coverage_90_lower = ifelse(
        .data[["output_type_id"]] == 0.05,
        .data[["oracle_value"]] >= .data[["value"]],
        NA_real_
      ),
      q_coverage_90_upper = ifelse(
        .data[["output_type_id"]] == 0.95,
        .data[["oracle_value"]] <= .data[["value"]],
        NA_real_
      )
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of(
      c(
        "model_id",
        "reference_date",
        "target",
        "horizon",
        "location",
        "target_end_date"
      )
    ))) |>
    dplyr::summarize(
      wis = 2 * mean(.data[["qs"]]),
      interval_coverage_80 = as.logical(
        (sum(.data[["q_coverage_80_lower"]], na.rm = TRUE) == 1) *
          (sum(.data[["q_coverage_80_upper"]], na.rm = TRUE) == 1)
      ),
      interval_coverage_90 = as.logical(
        (sum(.data[["q_coverage_90_lower"]], na.rm = TRUE) == 1) *
          (sum(.data[["q_coverage_90_upper"]], na.rm = TRUE) == 1)
      ),
      .groups = "drop"
    )

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  expect_equal(merged_scores$wis.x, merged_scores$wis.y)
  expect_equal(
    merged_scores$interval_coverage_80.x,
    merged_scores$interval_coverage_80.y
  )
  expect_equal(
    merged_scores$interval_coverage_90.x,
    merged_scores$interval_coverage_90.y
  )
})


test_that("score_model_out succeeds with valid inputs: nominal pmf output_type, default metrics, custom by", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "pmf"),
    oracle_output = forecast_oracle_output,
    by = c("model_id", "location")
  )

  exp_scores <- read.csv(test_path("testdata", "exp_pmf_scores.csv")) |>
    dplyr::mutate(location = as.character(location)) |>
    dplyr::select(-rps)

  # same column names, number of rows, and score values
  expect_equal(colnames(act_scores), colnames(exp_scores))
  expect_equal(nrow(act_scores), nrow(exp_scores))
  merged_scores <- dplyr::full_join(
    act_scores,
    exp_scores,
    by = c("model_id", "location")
  )
  expect_equal(nrow(act_scores), nrow(merged_scores))
  expect_equal(merged_scores$log_score.x, merged_scores$log_score.y)
})


test_that("score_model_out succeeds with valid inputs: ordinal pmf output_type, default metrics, custom by", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "pmf"),
    oracle_output = forecast_oracle_output,
    by = c("model_id", "location"),
    output_type_id_order = c("low", "moderate", "high", "very high")
  )

  exp_scores <- read.csv(test_path("testdata", "exp_pmf_scores.csv")) |>
    dplyr::mutate(location = as.character(location))

  # same answer
  expect_equal(act_scores, exp_scores, ignore_attr = TRUE)
})


test_that("score_model_out errors when model_out_tbl has multiple output_types", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs,
      oracle_output = forecast_oracle_output
    ),
    regexp = "model_out_tbl must contain a single output_type, but it has multiple"
  )
})


test_that("score_model_out works with all kinds of interval levels are requested", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = "interval_coverage_5d2a"
    ),
    regexp = "must be a number between 0 and 100"
  )

  suppressWarnings({
    # Non-standard interval level: scoringutils warns and produces no score
    # columns, then score_model_out() aborts (mirroring scoringutils#1180).
    expect_error(
      score_model_out(
        model_out_tbl = forecast_outputs |>
          dplyr::filter(.data[["output_type"]] == "quantile"),
        oracle_output = forecast_oracle_output,
        metrics = "interval_coverage_55"
      ),
      regexp = "No score columns to summarise"
    )

    expect_error(
      score_model_out(
        model_out_tbl = forecast_outputs |>
          dplyr::filter(.data[["output_type"]] == "quantile"),
        oracle_output = forecast_oracle_output,
        metrics = "interval_coverage_100"
      ),
      regexp = "must be a number between 0 and 100"
    )

    expect_error(
      score_model_out(
        model_out_tbl = forecast_outputs |>
          dplyr::filter(.data[["output_type"]] == "quantile"),
        oracle_output = forecast_oracle_output,
        metrics = "interval_coverage_5.3"
      ),
      regexp = "No score columns to summarise"
    )
  })
})


test_that("score_model_out errors when invalid metrics are requested", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = "log_score"
    ),
    regexp = "has additional elements"
  )

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = list(5, 6, "asdf")
    ),
    regexp = paste0(
      "^Assertion on 'c\\(select, exclude\\)' failed: Must be of",
      " type 'character' \\(or 'NULL'\\), not 'list'\\.$"
    )
  )

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      metrics = c("asdfinterval_coverage_90")
    ),
    regexp = "has additional elements"
  )

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "mean"),
      oracle_output = forecast_oracle_output,
      metrics = scoringutils::get_metrics(scoringutils::example_point),
      by = c("model_id", "location")
    ),
    regexp = paste0(
      "^Assertion on 'c\\(select, exclude\\)' failed: Must be of",
      " type 'character' \\(or 'NULL'\\), not 'list'\\.$"
    )
  )
})


test_that("score_model_out errors when an unsupported output_type is provided", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "cdf"),
      oracle_output = forecast_oracle_output,
      metrics = "log_score"
    ),
    regexp = "only supports the following types"
  )
})


# Tests for scale transformation functionality

test_that("transform argument produces same scores as manually transformed data", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_data <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Manually transform the data using sqrt
  quantile_data_transformed <- quantile_data |>
    dplyr::mutate(value = sqrt(.data[["value"]]))
  oracle_output_transformed <- forecast_oracle_output |>
    dplyr::mutate(oracle_value = sqrt(.data[["oracle_value"]]))

  # Score manually transformed data WITHOUT transform argument
  scores_manual <- score_model_out(
    model_out_tbl = quantile_data_transformed,
    oracle_output = oracle_output_transformed,
    metrics = "wis",
    by = "model_id"
  )

  # Score original data WITH transform argument
  scores_auto <- score_model_out(
    model_out_tbl = quantile_data,
    oracle_output = forecast_oracle_output,
    metrics = "wis",
    transform = sqrt,
    by = "model_id"
  )

  # Both approaches should produce identical scores

  expect_equal(scores_manual$wis, scores_auto$wis)
})


test_that("score_model_out succeeds with log transformation on median forecasts", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "median"),
    oracle_output = forecast_oracle_output,
    transform = scoringutils::log_shift,
    by = "model_id"
  )

  expect_true(nrow(scores) > 0)
  expect_true("ae_point" %in% colnames(scores))
})


test_that("score_model_out succeeds with log transformation on mean forecasts", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "mean"),
    oracle_output = forecast_oracle_output,
    transform = scoringutils::log_shift,
    by = "model_id"
  )

  expect_true(nrow(scores) > 0)
  expect_true("se_point" %in% colnames(scores))
})


test_that("score_model_out with transform_append=TRUE includes both scales", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "quantile"),
    oracle_output = forecast_oracle_output,
    metrics = "wis",
    transform = sqrt,
    transform_append = TRUE,
    summarize = FALSE
  )

  expect_true("scale" %in% colnames(scores))
  expect_true("natural" %in% scores$scale)
  expect_true("sqrt" %in% scores$scale)
})


test_that("score_model_out with transform_append=TRUE preserves scale when summarising (#122)", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output
  quantile_out <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Default summarize = TRUE + default by = "model_id" should keep the two
  # scales distinct (regression test: previously the natural and
  # transformed-scale rows were collapsed into a single average per model).
  scores <- score_model_out(
    model_out_tbl = quantile_out,
    oracle_output = forecast_oracle_output,
    metrics = "wis",
    transform = scoringutils::log_shift,
    transform_append = TRUE,
    offset = 1
  )

  expect_true("scale" %in% colnames(scores))
  expect_setequal(unique(scores$scale), c("natural", "log_shift"))
  expect_equal(nrow(scores), 2 * dplyr::n_distinct(scores$model_id))

  # Cross-check: default by must match explicit `by = c("model_id", "scale")`.
  scores_explicit <- score_model_out(
    model_out_tbl = quantile_out,
    oracle_output = forecast_oracle_output,
    metrics = "wis",
    transform = scoringutils::log_shift,
    transform_append = TRUE,
    offset = 1,
    by = c("model_id", "scale")
  )
  expect_equal(scores, scores_explicit)
})


test_that("score_model_out errors when transform requested for pmf output_type", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "pmf"),
      oracle_output = forecast_oracle_output,
      transform = scoringutils::log_shift
    ),
    regexp = "Scale transformations are not supported for pmf output types"
  )
})


test_that("score_model_out errors when transform is not a function", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      transform = "log"
    ),
    regexp = "transform.*must be a function or NULL"
  )
})


test_that("score_model_out errors when transform_label is not character", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      transform = scoringutils::log_shift,
      transform_label = 123
    ),
    regexp = "transform_label.*must be a character string or NULL"
  )
})


test_that("score_model_out passes ... arguments to transform function", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_data <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  # Using log_shift with offset=1 should avoid NaN scores and warnings
  # (data contains zeros which would cause issues without offset)
  scores <- score_model_out(
    model_out_tbl = quantile_data,
    oracle_output = forecast_oracle_output,
    metrics = "wis",
    transform = scoringutils::log_shift,
    by = "model_id",
    offset = 1
  )

  expect_true(nrow(scores) > 0)
  expect_false(any(is.nan(scores$wis)))
})


test_that("score_model_out errors when log_shift without offset produces Inf on zeros", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # Data contains zeros - log_shift without offset produces -Inf
  quantile_data <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")

  expect_error(
    suppressWarnings(score_model_out(
      model_out_tbl = quantile_data,
      oracle_output = forecast_oracle_output,
      metrics = "wis",
      transform = scoringutils::log_shift,
      by = "model_id"
    )),
    regexp = "non-finite values"
  )
})


# --- Sample scoring integration tests ---

test_that("score_model_out succeeds with sample output_type, marginal scoring, by model_id", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "sample"),
    oracle_output = forecast_oracle_output,
    metrics = "crps",
    by = "model_id"
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(scores, c("model_id", "crps"))
  expect_equal(nrow(scores), 3L)
})


test_that("score_model_out succeeds with sample output_type, summarize = FALSE", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "sample"),
    oracle_output = forecast_oracle_output,
    metrics = "crps",
    summarize = FALSE
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(
    scores,
    c(
      "model_id",
      "reference_date",
      "target",
      "horizon",
      "location",
      "target_end_date",
      "crps"
    )
  )
  # 3 models x 2 reference_dates x 2 locations x 4 horizons = 48
  expect_equal(nrow(scores), 48L)
})


test_that("score_model_out succeeds with sample output_type, default metrics", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # log_score uses kernel density estimation, which scoringutils correctly
  # warns is not appropriate for integer-valued forecasts
  expect_warning(
    scores <- score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "sample"),
      oracle_output = forecast_oracle_output,
      by = "model_id"
    ),
    regexp = "integer-valued"
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(
    scores,
    c(
      "model_id",
      "bias",
      "dss",
      "crps",
      "overprediction",
      "underprediction",
      "dispersion",
      "log_score",
      "mad",
      "ae_median",
      "se_mean"
    )
  )
  expect_equal(nrow(scores), 3L)
})


test_that("score_model_out computes correct CRPS for marginal sample scoring", {
  # Hand-computed CRPS for 3 samples {1, 3, 5}, observed = 2:
  # CRPS = (1/n) sum|x_i - y| - (1/(2n^2)) sum_ij|x_i - x_j|
  #      = (1/3)(1 + 1 + 3) - (1/18)(2 + 4 + 2 + 2 + 4 + 2) = 5/3 - 8/9 = 7/9
  model_out_tbl <- data.frame(
    model_id = "m1",
    output_type = "sample",
    output_type_id = as.character(1:3),
    value = c(1, 3, 5),
    location = "A",
    target = "inc hosp",
    target_end_date = as.Date("2024-01-01"),
    stringsAsFactors = FALSE
  )

  oracle_output <- data.frame(
    location = "A",
    target = "inc hosp",
    target_end_date = as.Date("2024-01-01"),
    oracle_value = 2,
    stringsAsFactors = FALSE
  )

  scores <- score_model_out(
    model_out_tbl = model_out_tbl,
    oracle_output = oracle_output,
    metrics = "crps",
    summarize = FALSE
  )

  expect_equal(scores$crps, 7 / 9)
})


test_that("score_model_out computes correct energy score for compound sample scoring", {
  # Hand-computed energy score for 3 bivariate samples over 2 horizons.
  # Samples: {(1,2), (3,4), (5,6)}, observed: (2,3)
  # ES = (1/n) sum||x_i - y|| - (1/(2n^2)) sum_ij||x_i - x_j||
  #    = (1/3)(sqrt(2) + sqrt(2) + 3*sqrt(2))
  #      - (1/18)(2*2*sqrt(2) + 2*4*sqrt(2) + 2*2*sqrt(2))
  #    = 5*sqrt(2)/3 - 8*sqrt(2)/9 = 7*sqrt(2)/9
  model_out_tbl <- data.frame(
    model_id = "m1",
    output_type = "sample",
    output_type_id = as.character(c(1, 1, 2, 2, 3, 3)),
    value = c(1, 2, 3, 4, 5, 6),
    location = "A",
    target = "inc hosp",
    horizon = c(1L, 2L, 1L, 2L, 1L, 2L),
    stringsAsFactors = FALSE
  )

  oracle_output <- data.frame(
    location = "A",
    target = "inc hosp",
    horizon = c(1L, 2L),
    oracle_value = c(2, 3),
    stringsAsFactors = FALSE
  )

  scores <- score_model_out(
    model_out_tbl = model_out_tbl,
    oracle_output = oracle_output,
    compound_taskid_set = "location",
    summarize = FALSE
  )

  expect_equal(scores$energy_score, 7 * sqrt(2) / 9)
})


test_that("score_model_out succeeds with compound sample scoring (energy score)", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  # compound_taskid_set matches the hub's tasks.json: reference_date and

  # location stay constant within each draw; horizon varies (joint_across).
  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "sample"),
    oracle_output = forecast_oracle_output,
    compound_taskid_set = c("reference_date", "location"),
    metrics = "energy_score",
    by = "model_id"
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(scores, c("model_id", "energy_score"))
  expect_equal(nrow(scores), 3L)
})


test_that("score_model_out succeeds with marginal sample and scale transformation", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  scores <- score_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output,
    metrics = "crps",
    transform = scoringutils::log_shift,
    offset = 1,
    by = "model_id"
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(scores, c("model_id", "crps"))
  expect_equal(nrow(scores), 3L)
})


test_that("score_model_out with sample transform_append=TRUE includes both scales", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  scores <- score_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output,
    metrics = "crps",
    transform = scoringutils::log_shift,
    offset = 1,
    transform_append = TRUE,
    summarize = FALSE
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(
    scores,
    c(
      "model_id",
      "reference_date",
      "target",
      "horizon",
      "location",
      "target_end_date",
      "scale",
      "crps"
    )
  )
  # 48 per scale x 2 scales = 96
  expect_equal(nrow(scores), 96L)
  expect_setequal(unique(scores$scale), c("natural", "log_shift"))
})


test_that("score_model_out succeeds with compound sample and scale transformation", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  scores <- score_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output,
    compound_taskid_set = c("reference_date", "location"),
    metrics = "energy_score",
    transform = scoringutils::log_shift,
    offset = 1,
    by = "model_id"
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(scores, c("model_id", "energy_score"))
  expect_equal(nrow(scores), 3L)
})


test_that("score_model_out with compound sample transform_append=TRUE includes both scales", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")

  scores <- score_model_out(
    model_out_tbl = sample_tbl,
    oracle_output = forecast_oracle_output,
    compound_taskid_set = c("reference_date", "location"),
    metrics = "energy_score",
    transform = scoringutils::log_shift,
    offset = 1,
    transform_append = TRUE,
    summarize = FALSE
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(
    scores,
    c(
      "model_id",
      "reference_date",
      "target",
      "location",
      "scale",
      "energy_score",
      ".mv_group_id"
    )
  )
  # 12 per scale (3 models x 2 ref_dates x 2 locations) x 2 scales = 24
  expect_equal(nrow(scores), 24L)
  expect_setequal(unique(scores$scale), c("natural", "log_shift"))
})


test_that("score_model_out succeeds with sample and relative metrics", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  scores <- score_model_out(
    model_out_tbl = forecast_outputs |>
      dplyr::filter(.data[["output_type"]] == "sample"),
    oracle_output = forecast_oracle_output,
    metrics = "crps",
    relative_metrics = "crps",
    by = "model_id"
  )

  expect_s3_class(scores, c("scores", "tbl_df", "tbl", "data.frame"))
  expect_named(scores, c("model_id", "crps", "crps_relative_skill"))
  expect_equal(nrow(scores), 3L)
})


test_that("score_model_out errors when compound_taskid_set used with non-sample type", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  expect_error(
    score_model_out(
      model_out_tbl = forecast_outputs |>
        dplyr::filter(.data[["output_type"]] == "quantile"),
      oracle_output = forecast_oracle_output,
      compound_taskid_set = c("reference_date", "location")
    ),
    regexp = "only applicable to sample output types"
  )
})


test_that("error_if_invalid_output_type accepts sample", {
  expect_no_error(error_if_invalid_output_type("sample"))
})


# --- Tests for non-finite transform validation (#99) ---

test_that("score_model_out errors with NaN when sqrt applied to negative sample values", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  sample_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample")
  # Inject a negative value to ensure sqrt produces NaN
  sample_tbl$value[1] <- -5

  expect_warning(
    expect_error(
      score_model_out(
        model_out_tbl = sample_tbl,
        oracle_output = forecast_oracle_output,
        metrics = "crps",
        transform = sqrt,
        by = "model_id"
      ),
      regexp = "non-finite values.*predicted.*NaN"
    ),
    regexp = "NaNs produced"
  )
})


test_that("score_model_out errors with Inf when log applied to zeros in quantile data", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  quantile_tbl <- forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile")
  # Inject a zero to ensure log produces -Inf
  quantile_tbl$value[1] <- 0

  expect_error(
    score_model_out(
      model_out_tbl = quantile_tbl,
      oracle_output = forecast_oracle_output,
      metrics = "wis",
      transform = log,
      by = "model_id"
    ),
    regexp = "non-finite values.*predicted.*Inf"
  )
})


test_that("score_model_out errors with NaN when log applied to negative sample values", {
  # Construct minimal sample data with a negative value
  model_out_tbl <- data.frame(
    model_id = "m1",
    output_type = "sample",
    output_type_id = as.character(1:3),
    value = c(-1, 3, 5),
    location = "A",
    target = "inc hosp",
    target_end_date = as.Date("2024-01-01"),
    stringsAsFactors = FALSE
  )

  oracle_output <- data.frame(
    location = "A",
    target = "inc hosp",
    target_end_date = as.Date("2024-01-01"),
    oracle_value = 2,
    stringsAsFactors = FALSE
  )

  expect_warning(
    expect_error(
      score_model_out(
        model_out_tbl = model_out_tbl,
        oracle_output = oracle_output,
        metrics = "crps",
        transform = log
      ),
      regexp = "non-finite values.*predicted.*NaN"
    ),
    regexp = "NaNs produced"
  )
})


test_that("score_model_out reports both predicted and observed when both have non-finite values", {
  # Both predicted and observed will have non-finite values from sqrt on negatives
  model_out_tbl <- data.frame(
    model_id = "m1",
    output_type = "sample",
    output_type_id = as.character(1:3),
    value = c(-1, 3, 5),
    location = "A",
    target = "inc hosp",
    target_end_date = as.Date("2024-01-01"),
    stringsAsFactors = FALSE
  )

  oracle_output <- data.frame(
    location = "A",
    target = "inc hosp",
    target_end_date = as.Date("2024-01-01"),
    oracle_value = -2,
    stringsAsFactors = FALSE
  )

  # sqrt is applied to both predicted and observed columns, so each
  # raises its own base-R "NaNs produced" warning -- hence two nested
  # expect_warning() calls before the hubEvals error is asserted.
  expect_warning(
    expect_warning(
      expect_error(
        score_model_out(
          model_out_tbl = model_out_tbl,
          oracle_output = oracle_output,
          metrics = "crps",
          transform = sqrt
        ),
        regexp = "predicted.*observed"
      ),
      regexp = "NaNs produced"
    ),
    regexp = "NaNs produced"
  )
})

library(dplyr)

# create model_out_tbl
model_out_tbl <- expand.grid(
  model_id = c("model_A", "model_B"),
  location = c("US", "01"),
  reference_date = as.Date(c("2020-01-14", "2020-01-21")),
  horizon = 0:2,
  output_type = "pmf",
  output_type_id = c("cat", "dog", "bird"),
  stringsAsFactors = FALSE
) |>
  dplyr::mutate(
    target_date = reference_date + 7 * horizon,
    value = row_number()
  ) |>
  dplyr::group_by(model_id, location, reference_date, target_date) |>
  dplyr::mutate(value = value / sum(value)) |>
  dplyr::ungroup()

# check that probs sum to 1
model_out_tbl |>
  group_by(model_id, location, reference_date, target_date) |>
  summarize(tot = sum(value)) |>
  pull(tot)

# create target_obserations
observed_categories <- data.frame(
  location = c("US", "US", "US", "US", "01", "01", "01", "01"),
  target_date = as.Date(c(
    "2020-01-14",
    "2020-01-21",
    "2020-01-28",
    "2020-02-04",
    "2020-01-14",
    "2020-01-21",
    "2020-01-28",
    "2020-02-04"
  )),
  output_type_id = c("cat", "dog", "cat", "bird", "bird", "dog", "dog", "cat"),
  observation = 1,
  stringsAsFactors = FALSE
)

oracle_output <- model_out_tbl |>
  dplyr::distinct(location, target_date, output_type_id) |>
  dplyr::left_join(
    observed_categories,
    by = c("location", "target_date", "output_type_id")
  ) |>
  dplyr::mutate(oracle_value = ifelse(is.na(observation), 0, 1)) |>
  dplyr::select(-observation)

# check that observations sum to 1
oracle_output |>
  group_by(location, target_date) |>
  summarize(tot = sum(oracle_value)) |>
  pull(tot)

# create expected forecast output
exp_forecast <- model_out_tbl |>
  dplyr::mutate(
    predicted_label = factor(output_type_id, levels = c("cat", "dog", "bird"))
  ) |>
  dplyr::select(-output_type, -output_type_id) |>
  dplyr::left_join(
    observed_categories |>
      dplyr::select(-observation) |>
      dplyr::rename(observed = output_type_id),
    by = c("location", "target_date")
  ) |>
  dplyr::mutate(
    observed = factor(observed, levels = c("cat", "dog", "bird"))
  ) |>
  dplyr::select(
    predicted_label,
    predicted = value,
    observed,
    model = model_id,
    location,
    reference_date,
    horizon,
    target_date
  )

class(exp_forecast) <- c(
  "forecast_nominal",
  "forecast",
  "data.table",
  "data.frame"
)


dput_fun <- function(obj, path = stdout()) {
  x <- get(obj)
  if (!is.function(x)) {
    x <- paste(capture.output(dput(x)), collapse = "\n  ")
    res <- glue::glue(
      "# nolint start\npmf_test_{obj} <- function() {{\n  {dput(x)}\n}}\n# nolint end"
    )
    writeLines(res, path)
  }
}
dput_fun("model_out_tbl", "tests/testthat/helper-pmf_test_model_out_tbl.R")
dput_fun("oracle_output", "tests/testthat/helper-pmf_test_oracle_output.R")
dput_fun(
  "exp_forecast",
  "tests/testthat/helper-pmf_test_expected_merged_forecast.R"
)

#' Geometric mean
#' (x_1 \times x_2 \times \ldots \times x_n)^{1/n}
#'  = exp[1/n \sum_{i=1}^{n} log(x_i)]
geometric_mean <- function(x) {
  exp(mean(log(x)))
}


#' Helper function manually computes pairwise relative skill scores by location
#' Called from tests in test-score_model_out_rel_metrics.R
get_pairwise_scores_by_loc <- function(scores_per_task, metric, baseline) {
  mean_scores_by_loc <- scores_per_task |>
    dplyr::group_by(dplyr::across(dplyr::all_of(c("model_id", "location")))) |>
    dplyr::summarize(
      mean_score = mean(.data[[metric]], na.rm = TRUE), # nolint: object_usage_linter
      .groups = "drop"
    )

  pairwise_score_ratios <- expand.grid(
    model_id = unique(mean_scores_by_loc$model_id),
    model_id_compare = unique(mean_scores_by_loc$model_id),
    location = unique(mean_scores_by_loc[["location"]])
  ) |>
    dplyr::left_join(
      mean_scores_by_loc,
      by = c("model_id" = "model_id", "location")
    ) |>
    dplyr::left_join(
      mean_scores_by_loc,
      by = c("model_id_compare" = "model_id", "location")
    ) |>
    dplyr::mutate(
      pairwise_score_ratio = .data[["mean_score.x"]] / .data[["mean_score.y"]] # nolint: object_usage_linter
    )

  result <- pairwise_score_ratios |>
    dplyr::group_by(dplyr::across(dplyr::all_of(c("model_id", "location")))) |>
    dplyr::summarize(
      relative_skill = geometric_mean(.data[["pairwise_score_ratio"]]), # nolint: object_usage_linter
      .groups = "drop"
    ) |>
    dplyr::group_by(dplyr::across(dplyr::all_of("location"))) |>
    dplyr::mutate(
      scaled_relative_skill = .data[["relative_skill"]] /
        .data[["relative_skill"]][.data[["model_id"]] == baseline]
    )

  colnames(result) <- c(
    "model_id",
    "location",
    paste0(metric, "_relative_skill"),
    paste0(metric, "_scaled_relative_skill")
  )

  result
}


# Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
forecast_outputs <- hubExamples::forecast_outputs
forecast_oracle_output <- hubExamples::forecast_oracle_output

# expected scores
exp_scores_unsummarized <- forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "quantile") |>
  dplyr::left_join(
    forecast_oracle_output |>
      dplyr::filter(.data[["output_type"]] == "quantile") |>
      dplyr::select(-dplyr::all_of(c("output_type", "output_type_id"))),
    by = c("location", "target_end_date", "target")
  ) |>
  dplyr::mutate(
    output_type_id = as.numeric(.data[["output_type_id"]]),
    qs = ifelse(
      .data[["oracle_value"]] >= .data[["value"]],
      .data[["output_type_id"]] * (.data[["oracle_value"]] - .data[["value"]]),
      (1 - .data[["output_type_id"]]) *
        (.data[["value"]] - .data[["oracle_value"]])
    ),
    q_cov_80_lower = ifelse(
      .data[["output_type_id"]] == 0.1,
      .data[["oracle_value"]] >= .data[["value"]],
      NA_real_
    ),
    q_coverage_80_upper = ifelse(
      .data[["output_type_id"]] == 0.9,
      .data[["oracle_value"]] <= .data[["value"]],
      NA_real_
    ),
    q_cov_90_lower = ifelse(
      .data[["output_type_id"]] == 0.05,
      .data[["oracle_value"]] >= .data[["value"]],
      NA_real_
    ),
    q_coverage_90_upper = ifelse(
      .data[["output_type_id"]] == 0.95,
      .data[["oracle_value"]] <= .data[["value"]],
      NA_real_
    )
  ) |>
  dplyr::group_by(dplyr::across(dplyr::all_of(
    c(
      "model_id",
      "location",
      "reference_date",
      "horizon",
      "target_end_date",
      "target"
    )
  ))) |>
  dplyr::summarize(
    ae_median = sum(ifelse(
      .data[["output_type_id"]] == 0.5,
      abs(.data[["oracle_value"]] - .data[["value"]]),
      0
    )),
    wis = 2 * mean(.data[["qs"]]),
    interval_coverage_80 = (sum(.data[["q_cov_80_lower"]], na.rm = TRUE) == 1) *
      (sum(.data[["q_coverage_80_upper"]], na.rm = TRUE) == 1),
    interval_coverage_90 = (sum(.data[["q_cov_90_lower"]], na.rm = TRUE) == 1) *
      (sum(.data[["q_coverage_90_upper"]], na.rm = TRUE) == 1)
  )

exp_scores_standard <- exp_scores_unsummarized |>
  dplyr::group_by(dplyr::across(dplyr::all_of(
    c("model_id", "location")
  ))) |>
  dplyr::summarize(
    ae_median = mean(.data[["ae_median"]]),
    wis = mean(.data[["wis"]]),
    interval_coverage_80 = mean(.data[["interval_coverage_80"]], na.rm = TRUE),
    interval_coverage_90 = mean(.data[["interval_coverage_90"]], na.rm = TRUE),
    .groups = "drop"
  )

# add pairwise relative scores for ae_median and wis
exp_scores_relative_ae_median <- get_pairwise_scores_by_loc(
  exp_scores_unsummarized,
  "ae_median",
  "Flusight-baseline"
)
exp_scores_relative_wis <- get_pairwise_scores_by_loc(
  exp_scores_unsummarized,
  "wis",
  "Flusight-baseline"
)
exp_scores <- exp_scores_standard |>
  dplyr::full_join(
    exp_scores_relative_ae_median,
    by = c("model_id", "location")
  ) |>
  dplyr::full_join(exp_scores_relative_wis, by = c("model_id", "location"))

# save
write.csv(
  exp_scores,
  testthat::test_path("testdata", "exp_pairwise_scores.csv"),
  row.names = FALSE
)

#' Transform quantile model output into a forecast object
#'
#' @param model_out_tbl Model output tibble with predictions
#' @param oracle_output Predictions that would have been generated by an oracle
#' model that knew the observed target data values in advance
#'
#' @return forecast_quantile
#'
#' @examplesIf requireNamespace("hubExamples", quietly = TRUE)
#' quantile_forecast <- hubExamples::forecast_outputs |>
#'   dplyr::filter(.data[["output_type"]] == "quantile") |>
#'   transform_quantile_model_out(
#'     oracle_output = hubExamples::forecast_oracle_output
#'   )
#' quantile_forecast
#'
#' @importFrom rlang .data
#' @export
transform_quantile_model_out <- function(model_out_tbl, oracle_output) {
  model_out_tbl <- validate_model_oracle_out(model_out_tbl, oracle_output)

  task_id_cols <- get_task_id_cols(model_out_tbl)

  model_out_tbl <- model_out_tbl |>
    dplyr::filter(.data[["output_type"]] == "quantile") |>
    dplyr::mutate(output_type_id = as.numeric(.data[["output_type_id"]])) |>
    dplyr::rename(model = "model_id")

  # Filter oracle_output to quantile rows only when it still carries the
  # `output_type` column (i.e. mixed output types).
  if ("output_type" %in% colnames(oracle_output)) {
    oracle_output <- dplyr::filter(
      oracle_output,
      .data[["output_type"]] == "quantile"
    )
  }
  # Always drop output_type/output_type_id if present: output_type_id can
  # remain after a user-side filter that removed output_type, and a stray
  # output_type_id breaks `as_forecast_quantile`.
  oracle_output <- dplyr::select(
    oracle_output,
    -dplyr::any_of(c("output_type", "output_type_id"))
  )

  data <- dplyr::left_join(
    model_out_tbl,
    oracle_output,
    by = task_id_cols[task_id_cols %in% colnames(oracle_output)],
    relationship = "many-to-one"
  )

  forecast_quantile <- scoringutils::as_forecast_quantile(
    data,
    forecast_unit = c("model", task_id_cols),
    observed = "oracle_value",
    predicted = "value",
    quantile_level = "output_type_id"
  )

  forecast_quantile
}

test_that("model_out_tbl_1 output is valid", {
  model_out_tbl_1 <- utils::read.csv(test_path(
    "testdata/model_out_tbl_quantile_1.csv"
  ))
  oracle_output_1 <- utils::read.csv(test_path("testdata/target_data_1.csv")) |>
    dplyr::rename(oracle_value = observation)
  act_forecast <- transform_quantile_model_out(
    model_out_tbl = model_out_tbl_1,
    oracle_output = oracle_output_1
  )

  exp_forecast <- utils::read.csv(test_path("testdata/exp_forecast_2.csv"))
  class(exp_forecast) <- c(
    "forecast_quantile",
    "forecast",
    "data.table",
    "data.frame"
  )
  expect_equal(act_forecast, exp_forecast)
})


test_that("model_out_tbl_1 columns are valid", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_quantile_1.csv")
  ) |>
    dplyr::rename(loc = location, trgt = target, date = target_end_date)

  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(
      loc = location,
      date = target_end_date,
      oracle_value = observation
    )
  act_forecast <- transform_quantile_model_out(
    model_out_tbl = model_out_tbl_1,
    oracle_output = oracle_output_1
  )
  exp_forecast <- utils::read.csv(
    test_path("testdata/exp_forecast_2.csv")
  ) |>
    dplyr::rename(loc = location, trgt = target, date = target_end_date)
  class(exp_forecast) <- c(
    "forecast_quantile",
    "forecast",
    "data.table",
    "data.frame"
  )
  expect_equal(act_forecast, exp_forecast, ignore_attr = "class")
})

test_that("model_out_tbl_1 has any rows", {
  # Error is thrown by scoringutils::assert_forecast_generic()
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_quantile_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(oracle_value = observation)
  expect_error(
    suppressWarnings(transform_quantile_model_out(
      model_out_tbl = model_out_tbl_1[0, ],
      oracle_output = oracle_output_1
    )),
    regexp = "Must have at least 1 rows, but has 0 rows."
  )
})


test_that("many-to-one relationship exists between model_out_tbl and oracle_output", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_quantile_1.csv")
  )
  oracle_output_2 <- utils::read.csv(
    test_path("testdata/target_data_2.csv")
  ) |>
    dplyr::rename(oracle_value = observation)

  expect_error(
    suppressMessages(transform_quantile_model_out(
      model_out_tbl = model_out_tbl_1,
      oracle_output = oracle_output_2
    )),
    regexp = "Each row in `x` must match at most 1 row in `y`."
  )
})

test_that("hubExamples data set is transformed correctly", {
  skip_if_not_installed("hubExamples")
  # Forecast data from hubExamples: <https://hubverse-org.github.io/hubExamples/reference/forecast_data.html>
  forecast_outputs <- hubExamples::forecast_outputs
  forecast_oracle_output <- hubExamples::forecast_oracle_output

  act_forecast <- transform_quantile_model_out(
    model_out_tbl = forecast_outputs,
    oracle_output = forecast_oracle_output
  )

  exp_forecast <- forecast_outputs |>
    dplyr::filter(output_type == "quantile") |>
    dplyr::rename(model = model_id, quantile_level = output_type_id) |>
    dplyr::mutate(quantile_level = as.numeric(quantile_level))
  expect_s3_class(
    act_forecast,
    c("forecast_quantile", "forecast", "data.table", "data.frame")
  )

  # same number of row as quantile predictions from model_out_tbl,
  # same predicted values
  expect_equal(nrow(exp_forecast), nrow(act_forecast))
  exp_act_forecast <- dplyr::full_join(
    exp_forecast,
    act_forecast,
    by = c(
      "model",
      "reference_date",
      "target",
      "horizon",
      "location",
      "target_end_date",
      "quantile_level"
    )
  )
  expect_equal(exp_act_forecast$predicted, exp_act_forecast$value)

  # correct observed values, in alignment with oracle_output
  exp_act_forecast <- dplyr::left_join(
    act_forecast,
    forecast_oracle_output |>
      dplyr::filter(output_type == "quantile") |>
      dplyr::rename(quantile_level = output_type_id) |>
      dplyr::mutate(quantile_level = as.numeric(quantile_level)),
    by = c("target", "location", "target_end_date")
  )
  expect_equal(exp_act_forecast$observed, exp_act_forecast$oracle_value)
})

test_that("oracle_output without output_type but with output_type_id works (#73)", {
  skip_if_not_installed("hubExamples")
  forecast_outputs <- hubExamples::forecast_outputs
  oracle_no_output_type <- hubExamples::forecast_oracle_output |>
    dplyr::filter(output_type == "quantile") |>
    dplyr::select(-"output_type")

  act_forecast <- transform_quantile_model_out(
    model_out_tbl = forecast_outputs,
    oracle_output = oracle_no_output_type
  )

  expect_s3_class(
    act_forecast,
    c("forecast_quantile", "forecast", "data.table", "data.frame")
  )

  # cross-check against the regular pathway (oracle keeps both columns)
  exp_forecast <- transform_quantile_model_out(
    model_out_tbl = forecast_outputs,
    oracle_output = hubExamples::forecast_oracle_output
  )
  expect_equal(act_forecast, exp_forecast)
})

# scoringutils' pairwise relative-skill calculation invokes `wilcox.test` on
# score pairs, which emits "cannot compute exact p-value with ties" (and the
# "...with zeroes" variant) whenever the score pairs tie or include zeroes. The
# hubExamples oracle ties heavily, so a relative-skill scoring pass surfaces
# several of these per run, cluttering test output without indicating any real
# problem. This helper muffles only those warnings, so any other warning still
# surfaces (and fails the test), unlike a blanket `suppressWarnings()`.
suppress_wilcox_ties_warnings <- function(expr) {
  withCallingHandlers(
    expr,
    warning = function(w) {
      if (grepl("cannot compute exact p-value", conditionMessage(w))) {
        invokeRestart("muffleWarning")
      }
    }
  )
}

#' Get the names of columns in model_out_tbl corresponding to task IDs
#'
#' @param model_out_tbl Model output tibble with predictions
#'
#' @return character vector of column names
#'
#' @noRd
get_task_id_cols <- function(model_out_tbl) {
  model_out_cols <- colnames(model_out_tbl)
  # To-do: hubUtils::std_colnames (non-task id cols)
  non_task_cols <- c(
    "model_id",
    "output_type",
    "output_type_id",
    "value",
    "model_abbr",
    "team_abbr"
  )
  task_id_cols <- model_out_cols[!model_out_cols %in% non_task_cols]

  task_id_cols
}

#' Validate model_out_tbl and oracle_output arguments to transform functions
#'
#' @param model_out_tbl Model output tibble with predictions
#' @param oracle_output Predictions that would have been generated by an oracle
#' model that knew the observed target data values in advance
#'
#' @return The input `model_out_tbl`, possibly modified to ensure it has S3 class `model_out_tbl`
#'
#' @noRd
validate_model_oracle_out <- function(model_out_tbl, oracle_output) {
  # check that: model_out_tbl contains columns: model_id, output_type, output_type_id, value
  req_cols <- c("model_id", "output_type", "output_type_id", "value")
  if (!all(req_cols %in% colnames(model_out_tbl))) {
    cli::cli_abort(
      "model_out_tbl does not contain required columns: model_id, output_type, output_type_id, value:
      {.val {colnames(model_out_tbl)}}"
    )
  }

  # check that model_out_tbl and oracle_output have compatible columns
  task_id_cols <- get_task_id_cols(model_out_tbl)
  if (length(task_id_cols[task_id_cols %in% colnames(oracle_output)]) == 0) {
    cli::cli_abort(
      "model_out_tbl and oracle_output do not have compatible columns"
    )
  }
  t_o_cols <- colnames(oracle_output)
  expected_cols_superset <- c(
    task_id_cols,
    "output_type",
    "output_type_id",
    "oracle_value"
  )
  unexpected_cols <- t_o_cols[!t_o_cols %in% expected_cols_superset]
  if (length(unexpected_cols) > 0) {
    cli::cli_abort(
      c(
        "`oracle_output` had {length(unexpected_cols)} unexpected column{?s} {.val {unexpected_cols}};",
        " expected the columns of `oracle_output` to be a subset of {.val {expected_cols_superset}}."
      )
    )
  }

  if (!c("oracle_value") %in% colnames(oracle_output)) {
    cli::cli_abort(
      "oracle_output does not have oracle_value column"
    )
  }

  if (!inherits(model_out_tbl, "model_out_tbl")) {
    model_out_tbl <- hubUtils::as_model_out_tbl(model_out_tbl)
  }

  model_out_tbl
}


#' Check that model_out_tble has a single `output_type` that is one of the
#' `output_types` that is supported by this function.
#'
#' @return if valid, the output_type in model_out_tbl
#'
#' @noRd
validate_output_type <- function(model_out_tbl) {
  output_type <- unique(model_out_tbl$output_type)
  if (length(output_type) != 1) {
    cli::cli_abort(
      "model_out_tbl must contain a single output_type, but it has multiple:
      {.val {output_type}}"
    )
  }

  error_if_invalid_output_type(output_type)

  output_type
}


error_if_invalid_output_type <- function(output_type) {
  supported_types <- c("mean", "median", "pmf", "quantile", "sample")
  if (!output_type %in% supported_types) {
    cli::cli_abort(
      "Provided `model_out_tbl` contains `output_type` {.val {output_type}};
      hubEvals currently only supports the following types:
      {.val {supported_types}}"
    )
  }
}


#' Validate relative metrics
#'
#' @noRd
validate_relative_metrics <- function(relative_metrics, metrics, by) {
  if (any(is_interval_coverage_metric(relative_metrics))) {
    cli::cli_abort(
      "Interval coverage metrics are not supported for relative skill scores."
    )
  }

  if ("bias" %in% relative_metrics) {
    cli::cli_abort(
      c(
        "{.val bias} is not supported for relative skill scores.",
        "i" = "Bias is a signed quantity, so a geometric-mean pairwise ratio has no clean interpretation.",
        "i" = "Drop {.val bias} from {.arg relative_metrics} (it can still be requested via {.arg metrics})."
      )
    )
  }

  if (length(relative_metrics) > 0 && !"model_id" %in% by) {
    cli::cli_abort(
      "Relative metrics require 'model_id' to be included in {.arg by}."
    )
  }

  extra_metrics <- setdiff(relative_metrics, metrics)
  if (length(extra_metrics) > 0) {
    cli::cli_abort(c(
      "Relative metrics must be a subset of the metrics.",
      "x" = "The following {.arg relative_metrics} are not in {.arg metrics}: {.val {extra_metrics}}"
    ))
  }
}


#' Infer transform label from function expression
#'
#' Extracts a label for the transformation function, either from user input
#' or by inferring it from the function name.
#'
#' @param transform_expr The captured expression for the transform argument,
#'   obtained via `rlang::enexpr(transform)` in the calling function.
#' @param transform_label User-provided label, or NULL to auto-generate
#' @param call The call to use in error messages (for cli::cli_abort)
#'
#' @return Character string label (e.g., "log_shift", "sqrt")
#'
#' @details
#' When `transform_label` is NULL, the function name is extracted from the
#' expression. For namespaced calls like `scoringutils::log_shift`, only the
#' function name ("log_shift") is used. Anonymous functions require an
#' explicit `transform_label`.
#'
#' @noRd
get_transform_label <- function(
  transform_expr,
  transform_label,
  call = rlang::caller_env()
) {
  if (!is.null(transform_label)) {
    return(transform_label)
  }

  # Check if transform was passed as a named function:
  # - is_symbol: bare function name like `sqrt` or `log_shift`
  # - is_call(, "::"): namespaced function like `scoringutils::log_shift`
  if (
    rlang::is_symbol(transform_expr) || rlang::is_call(transform_expr, "::")
  ) {
    # Extract function name, stripping namespace if present
    # e.g., "scoringutils::log_shift" -> "log_shift"
    return(sub(".*::", "", rlang::as_label(transform_expr)))
  }

  # Anonymous functions (e.g., function(x) log(x + 1)) don't have a name
  # to infer, so require explicit label
  cli::cli_abort(
    "{.arg transform_label} is required when using an anonymous transform function.",
    call = call
  )
}


#' Validate transformation parameters
#'
#' @param transform A transformation function or NULL
#' @param transform_label User-provided label or NULL
#' @param output_type The output_type from the model_out_tbl
#' @param call The call to use in error messages (for cli::cli_abort)
#'
#' @return TRUE (invisibly), or errors with informative message
#'
#' @noRd
validate_transform <- function(
  transform,
  transform_label,
  output_type,
  call = rlang::caller_env()
) {
  if (is.null(transform)) {
    return(invisible(TRUE))
  }

  if (!is.function(transform)) {
    cli::cli_abort(
      c(
        "{.arg transform} must be a function or NULL.",
        "x" = "Got {.cls {class(transform)}} instead."
      ),
      call = call
    )
  }

  if (!is.null(transform_label) && !is.character(transform_label)) {
    cli::cli_abort(
      c(
        "{.arg transform_label} must be a character string or NULL.",
        "x" = "Got {.cls {class(transform_label)}} instead."
      ),
      call = call
    )
  }

  if (output_type == "pmf") {
    cli::cli_abort(
      c(
        "Scale transformations are not supported for pmf output types.",
        "i" = "Transformations can only be applied to quantile, mean, median, or sample forecasts."
      ),
      call = call
    )
  }

  invisible(TRUE)
}


#' Validate transformed forecast values for non-finite entries
#'
#' Checks `predicted` and `observed` columns of a transformed scoringutils
#' forecast object for `NaN` and `Inf` values, which would silently corrupt
#' downstream scoring results.
#'
#' @param forecast A scoringutils forecast object (after `transform_forecasts()`).
#' @param call The call to use in error messages (for cli::cli_abort).
#'
#' @return `forecast` (invisibly), or errors with informative message.
#'
#' @noRd
validate_transform_values <- function(forecast, call = rlang::caller_env()) {
  cols_to_check <- intersect(c("predicted", "observed"), colnames(forecast))

  n_nan <- sum(purrr::map_int(cols_to_check, \(col) {
    sum(is.nan(forecast[[col]]))
  }))
  n_inf <- sum(purrr::map_int(cols_to_check, \(col) {
    sum(is.infinite(forecast[[col]]))
  }))

  if (n_nan == 0 && n_inf == 0) {
    return(invisible(forecast))
  }

  # nolint next: object_usage_linter
  affected_cols <- cols_to_check[purrr::map_lgl(
    cols_to_check,
    \(col) any(!is.finite(forecast[[col]]))
  )]

  cli::cli_abort(
    c(
      "Scale transformation produced non-finite values in
       {.field {affected_cols}}: {n_nan} NaN{?s} and {n_inf} Inf{?s}.",
      "x" = "Scoring results would be invalid with non-finite values.",
      "i" = "Use a transform that is valid for all values in the data."
    ),
    call = call
  )
}


#' Validate compound_taskid_set and check data consistency
#'
#' @param data Joined model output + oracle data, filtered to samples.
#' @param compound_taskid_set Character vector of task IDs that stay constant
#'   within each sample draw.
#' @param task_id_cols Character vector of all task ID column names.
#' @param call Calling environment for error messages.
#'
#' @noRd
validate_compound_taskid_set <- function(
  data,
  compound_taskid_set,
  task_id_cols,
  call = rlang::caller_env()
) {
  if (!is.character(compound_taskid_set) || length(compound_taskid_set) == 0) {
    cli::cli_abort(
      "{.arg compound_taskid_set} must be a non-empty character vector.",
      call = call
    )
  }

  invalid_cols <- setdiff(compound_taskid_set, task_id_cols)
  if (length(invalid_cols) > 0) {
    cli::cli_abort(
      c(
        "{.arg compound_taskid_set} contains invalid column{?s}: {.val {invalid_cols}}.",
        "i" = "Valid task ID columns: {.val {task_id_cols}}"
      ),
      call = call
    )
  }

  joint_across <- setdiff(task_id_cols, compound_taskid_set)
  if (length(joint_across) == 0) {
    cli::cli_abort(
      c(
        "{.arg compound_taskid_set} includes all task ID columns.",
        "i" = "At least one task ID column must be left out for joint scoring."
      ),
      call = call
    )
  }

  # A draw is identified by compound_taskid_set + output_type_id.
  # Sample IDs are only unique within each compound group, not globally.
  draw_group <- c(compound_taskid_set, "output_type_id")

  # Non-compound (joint_across) task IDs must vary within draws
  n_unique_per_draw <- data |>
    dplyr::group_by(dplyr::across(dplyr::all_of(draw_group))) |>
    dplyr::summarise(
      dplyr::across(
        dplyr::all_of(joint_across),
        dplyr::n_distinct
      ),
      .groups = "drop"
    )

  # Draws where ALL non-compound task IDs are constant (no joint structure)
  bad <- n_unique_per_draw |>
    dplyr::filter(dplyr::if_all(dplyr::all_of(joint_across), \(.x) .x == 1L))

  if (nrow(bad) > 0L) {
    bad_samples <- bad[["output_type_id"]] # nolint: object_usage_linter
    cli::cli_abort(
      c(
        "Non-compound task IDs must vary within sample draws for joint scoring,
         but {nrow(bad)} draw{?s} have no variation.",
        "i" = "First failing sample ID{?s}: {.val {utils::head(unique(bad_samples), 5)}}"
      ),
      call = call
    )
  }

  invisible(TRUE)
}

test_that("get_transform_label returns bare function name", {
  expr <- rlang::expr(sqrt)
  expect_equal(get_transform_label(expr, NULL), "sqrt")
})


test_that("get_transform_label strips namespace from namespaced function", {
  expr <- rlang::expr(scoringutils::log_shift)
  expect_equal(get_transform_label(expr, NULL), "log_shift")
})


test_that("get_transform_label returns user-provided label when given", {
  expr <- rlang::expr(sqrt)
  expect_equal(get_transform_label(expr, "my_label"), "my_label")

  # Also works with anonymous function expression when label provided
  expr_anon <- rlang::expr(function(x) log(x + 1))
  expect_equal(get_transform_label(expr_anon, "log1p"), "log1p")
})


test_that("get_transform_label errors for anonymous function without label", {
  expr <- rlang::expr(function(x) log(x + 1))
  expect_error(
    get_transform_label(expr, NULL),
    regexp = "transform_label.*required.*anonymous transform function"
  )
})


test_that("validate_relative_metrics rejects bias as a relative metric", {
  expect_error(
    validate_relative_metrics(
      relative_metrics = "bias",
      metrics = c("wis", "bias"),
      by = "model_id"
    ),
    regexp = "bias.*not supported for relative skill scores"
  )

  # Mixed with a valid relative metric: still errors
  expect_error(
    validate_relative_metrics(
      relative_metrics = c("wis", "bias"),
      metrics = c("wis", "bias"),
      by = "model_id"
    ),
    regexp = "bias.*not supported for relative skill scores"
  )
})

test_that("validate_model_oracle_out() throw errors for invalid inputs", {
  expect_error(
    validate_model_oracle_out(
      model_out_tbl = NULL,
      oracle_output = NULL
    ),
    regexp = "model_out_tbl does not contain required columns: model_id, output_type, output_type_id, value"
  )
  expect_error(
    suppressMessages(validate_model_oracle_out(
      model_out_tbl = data.frame(),
      oracle_output = NULL
    )),
    regexp = "model_out_tbl does not contain required columns: model_id, output_type, output_type_id, value"
  )
  expect_error(
    suppressMessages(validate_model_oracle_out(
      model_out_tbl = data.frame(),
      oracle_output = data.frame()
    )),
    regexp = "model_out_tbl does not contain required columns: model_id, output_type, output_type_id, value"
  )
})

test_that("valdiate_model_out_target_obs() works as expected for valid inputs", {
  # all clear
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(oracle_value = observation)
  val_result <- validate_model_oracle_out(
    model_out_tbl = model_out_tbl_1,
    oracle_output = oracle_output_1
  )

  # expected col names in order
  expected_cols <- c(
    "model_id",
    "location",
    "reference_date",
    "horizon",
    "target_end_date",
    "target",
    "model_abbr",
    "team_abbr",
    "output_type",
    "output_type_id",
    "value"
  )
  expect_equal(
    as.data.frame(val_result),
    model_out_tbl_1[, expected_cols]
  )
  expect_setequal(colnames(model_out_tbl_1), colnames(val_result))
})


test_that("an error is thrown if oracle_output is missing oracle_value column", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::select(-c("observation"))
  expect_error(
    suppressMessages(validate_model_oracle_out(
      model_out_tbl = model_out_tbl_1,
      oracle_output = oracle_output_1
    )),
    regexp = "oracle_output does not have oracle_value column"
  )
})


test_that("an error is thrown if model_out_tbl columns do not match oracle_output columns", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(oracle_value = observation)

  expect_error(
    suppressMessages(validate_model_oracle_out(
      model_out_tbl = model_out_tbl_1 |>
        dplyr::rename(loc = location, trgt = target, date = target_end_date),
      oracle_output = oracle_output_1
    )),
    regexp = "model_out_tbl and oracle_output do not have compatible columns"
  )
})


test_that("validate_model_oracle_out() throws error for unexpected columns", {
  model_out_tbl_1 <- utils::read.csv(
    test_path("testdata/model_out_tbl_point_1.csv")
  )
  oracle_output_1 <- utils::read.csv(
    test_path("testdata/target_data_1.csv")
  ) |>
    dplyr::rename(oracle_value = observation)
  expect_error(
    suppressMessages(validate_model_oracle_out(
      model_out_tbl = model_out_tbl_1 |>
        dplyr::rename(loc = location),
      oracle_output = oracle_output_1
    )),
    regexp = "unexpected column"
  )
})

# Helper used in inline roxygen `r ` expressions to generate bulleted metric
# lists without exceeding the line length limit.

#' @noRd
.metrics_list <- function(
  forecast_example = NULL,
  class = NULL,
  exclude = NULL
) {
  if (!is.null(forecast_example)) {
    metrics <- scoringutils::get_metrics(forecast_example, exclude = exclude)
  } else {
    fn <- utils::getS3method(
      "get_metrics",
      class,
      envir = asNamespace("scoringutils")
    )
    metrics <- fn()
  }
  paste("- ", names(metrics), collapse = "\n")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_pmf_model_out.R
\name{transform_pmf_model_out}
\alias{transform_pmf_model_out}
\title{Transform pmf model output into a forecast object}
\usage{
transform_pmf_model_out(
  model_out_tbl,
  oracle_output,
  output_type_id_order = NULL
)
}
\arguments{
\item{model_out_tbl}{Model output tibble with predictions}

\item{oracle_output}{Predictions that would have been generated by an oracle
model that knew the observed target data values in advance}

\item{output_type_id_order}{For ordinal variables in pmf format, this is a
vector of levels for pmf forecasts, in increasing order of the levels. The
order of the values for the output_type_id can be found by referencing the
hub's tasks.json configuration file. For all output types other than pmf,
this is ignored.}
}
\value{
A \code{forecast_ordinal} object (when \code{output_type_id_order} is provided)
or a \code{forecast_nominal} object (when \code{output_type_id_order} is \code{NULL}).
}
\description{
Transform pmf model output into a forecast object
}
\examples{
\dontshow{if (requireNamespace("hubExamples", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Nominal pmf forecast (no output_type_id_order provided)
pmf_outputs <- hubExamples::forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "pmf")

nominal_forecast <- transform_pmf_model_out(
  model_out_tbl = pmf_outputs,
  oracle_output = hubExamples::forecast_oracle_output
)
nominal_forecast

# Ordinal pmf forecast (output_type_id_order provided)
ordinal_forecast <- transform_pmf_model_out(
  model_out_tbl = pmf_outputs,
  oracle_output = hubExamples::forecast_oracle_output,
  output_type_id_order = c("low", "moderate", "high", "very high")
)
ordinal_forecast
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_quantile_model_out.R
\name{transform_quantile_model_out}
\alias{transform_quantile_model_out}
\title{Transform quantile model output into a forecast object}
\usage{
transform_quantile_model_out(model_out_tbl, oracle_output)
}
\arguments{
\item{model_out_tbl}{Model output tibble with predictions}

\item{oracle_output}{Predictions that would have been generated by an oracle
model that knew the observed target data values in advance}
}
\value{
forecast_quantile
}
\description{
Transform quantile model output into a forecast object
}
\examples{
\dontshow{if (requireNamespace("hubExamples", quietly = TRUE)) withAutoprint(\{ # examplesIf}
quantile_forecast <- hubExamples::forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "quantile") |>
  transform_quantile_model_out(
    oracle_output = hubExamples::forecast_oracle_output
  )
quantile_forecast
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_sample_model_out.R
\name{transform_sample_model_out}
\alias{transform_sample_model_out}
\title{Transform sample model output into a forecast object}
\usage{
transform_sample_model_out(
  model_out_tbl,
  oracle_output,
  compound_taskid_set = NULL
)
}
\arguments{
\item{model_out_tbl}{Model output tibble with predictions}

\item{oracle_output}{Predictions that would have been generated by an oracle
model that knew the observed target data values in advance}

\item{compound_taskid_set}{Character vector of task ID column names that
stay constant within each sample draw (i.e., define the compound modeling
task grouping). When \code{NULL} (the default), each modeling task is scored
independently (marginal scoring). When provided, sample draws are treated
as joint predictions across the task ID dimensions \strong{not} in
\code{compound_taskid_set}, and multivariate scoring metrics are used.}
}
\value{
A \code{forecast_sample} object (when \code{compound_taskid_set} is \code{NULL})
or a \code{forecast_multivariate_sample} object (when \code{compound_taskid_set} is
provided).
}
\description{
Transform sample model output into a forecast object
}
\examples{
\dontshow{if (requireNamespace("hubExamples", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Marginal sample forecast: each modeling task scored independently
sample_forecast <- hubExamples::forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "sample") |>
  transform_sample_model_out(
    oracle_output = hubExamples::forecast_oracle_output
  )
sample_forecast

# Compound sample forecast: jointly score across non-compound task IDs
compound_forecast <- hubExamples::forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "sample") |>
  transform_sample_model_out(
    oracle_output = hubExamples::forecast_oracle_output,
    compound_taskid_set = c("reference_date", "location")
  )
compound_forecast
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_point_model_out.R
\name{transform_point_model_out}
\alias{transform_point_model_out}
\title{Transform either mean or median model output into a point forecast object:}
\usage{
transform_point_model_out(model_out_tbl, oracle_output, output_type)
}
\arguments{
\item{model_out_tbl}{Model output tibble with predictions}

\item{oracle_output}{Predictions that would have been generated by an oracle
model that knew the observed target data values in advance}

\item{output_type}{Forecast output type: "mean" or "median"}
}
\value{
forecast_point
}
\description{
Transform either mean or median model output into a point forecast object:
}
\details{
This function transforms a model output tibble in the Hubverse
format (with either "mean" or "median" output type) to a scoringutils "point"
forecast object
}
\examples{
\dontshow{if (requireNamespace("hubExamples", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# Median point forecast
median_forecast <- hubExamples::forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "median") |>
  transform_point_model_out(
    oracle_output = hubExamples::forecast_oracle_output,
    output_type = "median"
  )
median_forecast

# Mean point forecast
mean_forecast <- hubExamples::forecast_outputs |>
  dplyr::filter(.data[["output_type"]] == "mean") |>
  transform_point_model_out(
    oracle_output = hubExamples::forecast_oracle_output,
    output_type = "mean"
  )
mean_forecast
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/score_model_out.R
\name{score_model_out}
\alias{score_model_out}
\title{Score model output predictions}
\usage{
score_model_out(
  model_out_tbl,
  oracle_output,
  metrics = NULL,
  relative_metrics = NULL,
  baseline = NULL,
  summarize = TRUE,
  by = "model_id",
  output_type_id_order = NULL,
  compound_taskid_set = NULL,
  transform = NULL,
  transform_append = FALSE,
  transform_label = NULL,
  ...
)
}
\arguments{
\item{model_out_tbl}{Model output tibble with predictions}

\item{oracle_output}{Predictions that would have been generated by an oracle
model that knew the observed target data values in advance}

\item{metrics}{Character vector of scoring metrics to compute. If \code{NULL}
(the default), appropriate metrics are chosen automatically. See details
for more.}

\item{relative_metrics}{Character vector of scoring metrics for which to
compute relative skill scores. The \code{relative_metrics} should be a subset of
\code{metrics} and should only include proper scores. Interval coverage metrics
(e.g., \code{"interval_coverage_90"}) and \code{"bias"} are not supported as relative
metrics: interval coverage targets a nominal level rather than a "lower is
better" direction, and bias is a signed quantity for which a geometric-mean
pairwise ratio has no clean interpretation. If \code{NULL} (the default), no
relative metrics will be computed. Relative metrics are only computed if
\code{summarize = TRUE}, and require that \code{"model_id"} is included in \code{by}.
Relative skill is computed within each comparison group (the \code{by} columns
other than \code{"model_id"}). If a comparison group contains only one model,
that group's relative-skill columns are filled with \code{1} (a model has trivial
skill \code{1} relative to itself) rather than erroring.}

\item{baseline}{String with the name of a model to use as a baseline for
relative skill scores. If a baseline is given, then a scaled relative skill
with respect to the baseline will be returned. By default (\code{NULL}), relative
skill will not be scaled with respect to a baseline model. If the baseline is
absent from a comparison group, that group's relative and scaled relative
skill are reported as \code{NA} (with a warning) while its absolute scores are
still returned. This \code{NA} case only arises when scoring is disaggregated:
with \code{by = "model_id"} alone there is a single comparison group containing
every model, so relative skill is always defined and is never \code{NA}. A
baseline absent from the data entirely is an error.}

\item{summarize}{Boolean indicator of whether summaries of forecast scores
should be computed. Defaults to \code{TRUE}.}

\item{by}{Character vector naming columns to summarize by. For example,
specifying \code{by = "model_id"} (the default) will compute average scores for
each model.}

\item{output_type_id_order}{For ordinal variables in pmf format, this is a
vector of levels for pmf forecasts, in increasing order of the levels. The
order of the values for the output_type_id can be found by referencing the
hub's tasks.json configuration file. For all output types other than pmf,
this is ignored.}

\item{compound_taskid_set}{When \code{NULL} (the default), sample forecasts are
scored marginally (each modeling task scored independently). When a character
vector of task ID column names is provided, it sets the compound grouping:
the specified columns stay constant within each sample draw, while the
remaining task ID dimensions vary within a draw and are scored jointly
(using the energy score). Only applicable when
\code{output_type == "sample"}. The value of \code{compound_taskid_set} can be found
by referencing the hub's \code{tasks.json} configuration file.}

\item{transform}{A function to apply as a scale transformation to both
predictions and observations before scoring. Common choices include
\code{\link[scoringutils]{log_shift}} (recommended for log transformation
as it handles zeros via an offset parameter), \code{sqrt}, or
\code{log1p}. Avoid using \code{log} directly if data may contain zeros.
If \code{NULL} (the default), no transformation is applied. Only supported for
quantile, mean, and median output types.}

\item{transform_append}{Logical. If \code{FALSE} (the default), scores are
computed only on the transformed scale. If \code{TRUE}, scores are computed on
both original and transformed scales, with a \code{scale} column distinguishing
them. Ignored if \code{transform = NULL}.}

\item{transform_label}{A character string label for the transformation
(e.g., "log"). If \code{NULL} (the default), the label is auto-generated from
the function name (e.g., "log_shift" for \code{scoringutils::log_shift}).
Required when using an anonymous transform function. Ignored if
\code{transform = NULL}. Note: the label only appears in output when
\code{transform_append = TRUE}, where it distinguishes transformed rows
(labeled with this value) from original rows (labeled "natural") in the
\code{scale} column.}

\item{...}{Additional arguments passed to the \code{transform} function. For
example, allows use of the \code{offset} and \code{base} arguments of
\code{\link[scoringutils:log_shift]{scoringutils::log_shift()}}. Ignored if \code{transform = NULL}.}
}
\value{
A tibble of scores, inheriting from scoringutils' \code{scores} class so
that downstream scoringutils helpers (e.g. \code{\link[scoringutils:get_metrics]{scoringutils::get_metrics()}})
continue to work. The tibble has a \code{metrics} attribute holding the names of
the scoring rules that were applied.
}
\description{
Scores model outputs with a single \code{output_type} against observed data.
}
\details{
See the hubverse documentation for the expected format of the
\href{https://docs.hubverse.io/en/latest/user-guide/target-data.html#oracle-output}{oracle output data}.

Default metrics are provided by the \code{scoringutils} package. You can select
metrics by passing in a character vector of metric names to the \code{metrics}
argument.

The following metrics can be selected (all are used by default) for the
different \code{output_type}s:

\strong{Quantile forecasts:} (\code{output_type == "quantile"})
\itemize{
\item wis
\item overprediction
\item underprediction
\item dispersion
\item bias
\item ae_median
\item "interval_coverage_XX": interval coverage at the "XX" level. For example,
"interval_coverage_95" is the 95\% interval coverage rate, which would be calculated
based on quantiles at the probability levels 0.025 and 0.975.
}

See \link[scoringutils:get_metrics.forecast_quantile]{scoringutils::get_metrics.forecast_quantile} for details.

\strong{Nominal forecasts:} (\code{output_type == "pmf"} and \code{output_type_id_order} is \code{NULL})
\itemize{
\item log_score
}

See \link[scoringutils:get_metrics.forecast_nominal]{scoringutils::get_metrics.forecast_nominal} for details.

\strong{Ordinal forecasts:} (\code{output_type == "pmf"} and \code{output_type_id_order} is a vector)
\itemize{
\item log_score
\item rps
}

See \link[scoringutils:get_metrics.forecast_ordinal]{scoringutils::get_metrics.forecast_ordinal} for details.

\strong{Median forecasts:} (\code{output_type == "median"})
\itemize{
\item ae_point: absolute error of the point forecast (recommended for the median, see Gneiting (2011))
}

See \link[scoringutils:get_metrics.forecast_point]{scoringutils::get_metrics.forecast_point} for details.

\strong{Mean forecasts:} (\code{output_type == "mean"})
\itemize{
\item \code{se_point}: squared error of the point forecast (recommended for the mean, see Gneiting (2011))
}

\strong{Sample forecasts (marginal):} (\code{output_type == "sample"}, \code{compound_taskid_set = NULL})
\itemize{
\item bias
\item dss
\item crps
\item overprediction
\item underprediction
\item dispersion
\item log_score
\item mad
\item ae_median
\item se_mean
}

Note: \code{log_score} uses kernel density estimation, which may not be
appropriate for integer-valued forecasts. \code{scoringutils} will warn when
this is detected.

See \link[scoringutils:get_metrics.forecast_sample]{scoringutils::get_metrics.forecast_sample} for details.

\strong{Sample forecasts (compound):} (\code{output_type == "sample"}, \code{compound_taskid_set} provided)
\itemize{
\item energy_score
\item variogram_score
}

See \link[scoringutils:get_metrics.forecast_multivariate_sample]{scoringutils::get_metrics.forecast_multivariate_sample} for details.
The output includes a \code{.mv_group_id} column assigned by \code{scoringutils} to
identify the multivariate groups used for scoring (equivalent to the
\code{compound_idx} concept in the
\href{https://docs.hubverse.io/en/latest/user-guide/sample-output-type.html}{hubverse sample output type documentation }).
Correct scoring depends on providing the right \code{compound_taskid_set} from
the hub's \code{tasks.json} configuration. If the specified grouping does not
match the actual dependence structure of the submitted samples (e.g.,
because some models submitted coarser samples than configured),
\code{.mv_group_id} may not correspond to the original sample draws as
indicated by their \code{output_type_id} values.

See \link[scoringutils:add_relative_skill]{scoringutils::add_relative_skill} for details on relative skill scores.
}
\examples{
\dontshow{if (requireNamespace("hubExamples", quietly = TRUE)) withAutoprint(\{ # examplesIf}
# compute WIS and interval coverage rates at 80\% and 90\% levels based on
# quantile forecasts, summarized by the mean score for each model
quantile_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "quantile"),
  oracle_output = hubExamples::forecast_oracle_output,
  metrics = c("wis", "interval_coverage_80", "interval_coverage_90"),
  relative_metrics = "wis",
  by = "model_id"
)
quantile_scores

# compute log scores based on pmf predictions for categorical targets,
# summarized by the mean score for each combination of model and location.
# Note: if the model_out_tbl had forecasts for multiple targets using a
# pmf output_type with different bins, it would be necessary to score the
# predictions for those targets separately.
pmf_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "pmf"),
  oracle_output = hubExamples::forecast_oracle_output,
  metrics = c("log_score", "rps"),
  by = c("model_id", "location", "horizon"),
  output_type_id_order = c("low", "moderate", "high", "very high")
)
head(pmf_scores)

# Score sample forecasts marginally (each modeling task scored independently).
# Note: this data has compound structure (samples span horizons), but marginal
# scoring is still valid -- it evaluates each horizon independently.
sample_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample"),
  oracle_output = hubExamples::forecast_oracle_output,
  metrics = "crps",
  by = "model_id"
)
sample_scores

# Score compound sample forecasts jointly using the energy score.
# compound_taskid_set specifies which task IDs stay constant within
# a sample group -- here, each sample draw spans all horizons for a
# given reference_date and location (i.e., a trajectory over time).
compound_scores <- score_model_out(
  model_out_tbl = hubExamples::forecast_outputs |>
    dplyr::filter(.data[["output_type"]] == "sample"),
  oracle_output = hubExamples::forecast_oracle_output,
  compound_taskid_set = c("reference_date", "location"),
  by = "model_id"
)
compound_scores
\dontshow{\}) # examplesIf}
}
\references{
Gneiting, Tilmann. 2011. "Making and Evaluating Point Forecasts." Journal of the
American Statistical Association 106 (494): 746–62. \doi{10.1198/jasa.2011.r10138}.
}
